// hvr_work.cpp : Implementation of Chvr_work
////////////////////////////////////////////////////////////////////////////////
//
// Author:   David H. Porter
//           Laboratory for Computational Science & Engineering
//           University of Minnesota
//
////////////////////////////////////////////////////////////////////////////////
/***************************************************************************************
Program:  HVR_SERVER  --  Hierarchical Volume Rendering Component Object Moduel
Copyright (C) 2002  David H. Porter

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
***************************************************************************************/

#include "stdafx.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#include <sys/timeb.h>

#include <gl\gl.h>
//#include "C:\include\GL\gl3dlabsext.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <process.h>

// rendparams.h is copied into in hvr_server.h    #include "renderparams.h"
#include "Hvr_server.h"
#include "hvr_work.h"
#include "TimeStamp.h"

//  ine DHP_TEX GL_COLOR_INDEX8_EXT
#define DHP_TEX GL_TEXTURE_2D



/////////////////////////////////////////////////////////////////////////////
// Chvr_work
//        *******************************************************
//        *******************************************************
//        *******************************************************
//        *******************************************************
//        *******************************************************
//        *******************************************************
//        *******************************************************
//        *******************************************************

static SIZE_T dwMaxSize = 128*128*128;
static unsigned char *bobsubX;
static unsigned char *bobsubY;
static unsigned char *bobsubZ;
static unsigned char *bobsubZ0;
static unsigned char *bobOX;
static unsigned char *bobOY;
static unsigned char *bobOZ;
static bool bBobsAllocated = FALSE;
static unsigned char sob[256*256];
static unsigned char ucGPix[3*1600*1200];
static unsigned char ucOPix[3*1600*64];
static long sol[512*512];
static bool bSetBobX, bSetBobY;
static HvrInput global_ShvrIn;			// GUI writes input here

static    HDC hDC;					/* device context */
static    HGLRC hRC;				/* opengl context */
static    HWND  hWnd;				/* window */
static    MSG   msg;				/* message */
static    int iFirst_Test = 1;
static    int iGLwindowIsMade = 0;
static    int iNClients = 0;
static    long iNRendering = 0;
static    int iAmUp = 0;
static    int iStopRendering = 0;
static    int iDoThis = 0;
static    wchar_t wcKey[256], wcValue[256];

CTimeStamp tsEvent;

typedef void (APIENTRY *PFNGLCOLORTABLEPROC)
	(GLenum target, GLenum internalformat, GLsizei width,
	GLenum format, GLenum type, const GLvoid *data );
PFNGLCOLORTABLEPROC glColorTableEXT;

// PFNGLGETCOLORTABLEPARAMETERIVEXTPROC glGetColorTableParameterivEXT;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Chvr_work::Chvr_work()
{
//	if(iAmUp == 1) { return; }
	iAmUp = 1;

	int i, j;
    void  setstr(char*, char*);

	isubcur = -1;         /* current sub-brick loaded into bobsub */
	ireload = 1;          /* flag to reload brick in 3dtex code   */
	Pid180 = (float)(3.141592654/180.);
	m_flClearRGB[0] = 0.;
	m_flClearRGB[1] = 0.;
	m_flClearRGB[2] = 0.;
	m_iRender = 0;
	m_ShvrIn.iNewHV      = 1;
	m_ShvrIn.iNewLUT     = 1;
	m_ShvrIn.iNewView    = 1;
	m_ShvrIn.iNewFrustum = 1;
	sprintf(m_sTimerFile, "HVR_timers.pto");
	sprintf(m_sStats, "none");
	m_iTimersOn = 0;
	m_bColorTableIsAvailable = FALSE;
	m_iStopRendering = 0;
	m_iBufferedIO = 0;
	m_iNRendBuffs = 2;
	m_iShowColorBar = 0;
	m_iNTicks = 0;
	m_iXNTicks = 0;
	m_iYNTicks = 0;
	iStopRendering = 0;
	m_iBlockOnRender = 1;
/*
	for(i=0; i<255; i += 50) {
		m_iTicks[m_iNTicks] = i;
		sprintf(m_sTicks[m_iNTicks], "%4.2f", (float)i * (float)(1.0/256.));
		if(i == 100) {
			sprintf(m_sTicks[m_iNTicks], "%4.2f=W[m/s]", (float)i * (float)(1.0/256.));
		}
		m_iNTicks++;
	}
*/
	SetDefaultFrustum();
	SetDefaultView(); 

	setstr(m_ShvrIn.sCmapname, "default");
	setstr(m_ShvrIn.sAmapname, "default");
	sprintf(m_cOldHVname, "none");

	for(i=0; i<256; i++) {
		for(j=0; j<4; j++) { frgba[i][j] = (float)i; }
		m_lrgba[i] = i + 256*(i + 256*i + 256*i);
	}
	m_iNrgbNots = 2;
	m_iNalpNots = 2;
	m_iRGBind[0] = 0;
	m_iALPind[0] = 0;
	m_fRED[0]    = (float)0.;
	m_fGRN[0]    = (float)0.;
	m_fBLU[0]    = (float)0.;
	m_fALP[0]    = (float)0.;
	m_iRGBind[1] = 255;
	m_iALPind[1] = 255;
	m_fRED[1]    = (float)1.;
	m_fGRN[1]    = (float)1.;
	m_fBLU[1]    = (float)1.;
	m_fALP[1]    = (float)1.;

	opacity   = 1.0;
	tolerance = 5;				/* max color level error allowed with no scan down  */
	angtol    = (float)0.02;	/* max angle allowed with no scan down              */
	viewres   = (float)8.0;		/* dx_brick > vrtolerance*dx_irep1 => no scan down  */
	nplanes   = 1;
	maxplanes = 1;

	m_ShvrIn.opacity = opacity;
	m_ShvrIn.tolerance = tolerance;
	m_ShvrIn.angtol = angtol;
	m_ShvrIn.viewres = viewres;
	m_ShvrIn.nplanes = nplanes;
	m_ShvrIn.maxplanes = maxplanes;


	pannels_initted = 0;
	softdistance = 0.;
	m_hLocalWnd = NULL;
	setstr(m_cOldCmap, "none");
	m_iAlphaIsSet = 0;

	if(!bBobsAllocated) {
		bobsubX  = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
		bobsubY  = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
		bobsubZ0 = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
		bobOX    = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
		bobOY    = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
		bobOZ    = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
		bBobsAllocated = TRUE;
	}
}

/* do I need this destructor here?
Chvr_work::~Chvr_work()
{

}
*/

STDMETHODIMP Chvr_work::ClearIt()
{
	// Only one thread can do any of the OpenGL calls, including shut down.
	// Hence, set flag iDoThis so that OpenGL thread will shut down.
	// See routine SpawnWork
	iDoThis = 2;
	while(iDoThis == 2) { _sleep(10); }

	return S_OK;
}

/*
LPVOID pParam;
*/
static int iOpenGL_WindowIsOpenned = 0;
static int iIO_Thread_Is_Up = 0;
static int iIO_Read = 0;
static int iN_IO_Threads=1;
static long i_IO_Thread=0;
static int iThreadIsOnCache[100];

void OpenAWindow(void *pParam)
{
	// Only do the body of this function ONCE per instance of the host program
//	if(iOpenGL_WindowIsOpenned == 1) return;
//	iOpenGL_WindowIsOpenned = 1;
	if(iGLwindowIsMade == 1) return;

	Chvr_work* my_work = (Chvr_work*)pParam;

	my_work->SpawnWork();
}

void StartIOThread(void *pParam)
{
	// Only do the body of this function ONCE per instance of the host program
//	if(iOpenGL_WindowIsOpenned == 1) return;
//	iOpenGL_WindowIsOpenned = 1;
	if(iIO_Thread_Is_Up == 1) return;

	Chvr_work* my_work = (Chvr_work*)pParam;

	my_work->DoIO();
}

static int     iNRequests;
static int     iCacheStatusIsReady;
static int     iTile_of_iReq[8192];
static int     iSorted_Req_of_iReq[8192];
static char    cFileName_of_iReq[8192][256];
static __int64 i64FileOffset_of_iReq[8192];
static DWORD   dwSize_of_iReq[8192];

static int            iNCacheMB = 160;
static int            iTile_of_iCache[8192];
static int            iUsed_of_iCache[8192];
static unsigned char  *lpBuffer_of_iCache[8192];
static long           lBufferMin_iCache[8192];
static long           lBufferMax_iCache[8192];
static int            iCache_of_iTile[8192];
static unsigned char  *BobCache;

void ResetCacheStatus()
{
	iCacheStatusIsReady = 0;
	iNRequests = 0;
}

void FlushCacheBuffers()
{
	int i;

	// Invalidate all cache buffers
	for(i=0; i<8192; i++) {
		iTile_of_iCache[i] = -1;
		iCache_of_iTile[i] = -1;
		iUsed_of_iCache[i] =  0;
	}
}

void RequestCacheIO(int iTile, LPCTSTR lpFileName, __int64 offset, DWORD nbytes)
{
	iTile_of_iReq[iNRequests] = iTile;
	sprintf(cFileName_of_iReq[iNRequests], "%s", lpFileName);
	i64FileOffset_of_iReq[iNRequests] = offset;
	dwSize_of_iReq[iNRequests] = nbytes;
	iNRequests++;
}

LPVOID lpGetCacheBuffer(int iTile)
{
	// Wait for cache status arrays to be ready
	while(iCacheStatusIsReady == 0   &&  iIO_Read == 1) { _sleep(10); }

	// Wait for cache buffer to be read in
	while(iCache_of_iTile[iTile] < 0) {
		if(iStopRendering == 1) { return BobCache; }
		_sleep(10);
	}

	// Return pointer to cache buffer
	return lpBuffer_of_iCache[iCache_of_iTile[iTile]];
}


void Chvr_work::DoIO()
{
	int i, ij, iWrap, isHere, isFree, iOverlap, iTile, iRequest, jRequest, iPrinted;
	int iCache, iCache0;
	int iPreviousOverFlow = 0, iMyThread, iNConsecutive, iCacheMin;
	int iTileTmp[128], iNSort, iMax, iChunk, iReqMin;
	long lBufferOffset, lNextBufferOffset, lNeedTotalBufferSize, lUnit;
	char cFileName[256];
	DWORD  dhpReadFromFileSync(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes);
	DWORD dhpReadFromFileNoBuf(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes);
//	DWORD dhpReadFromOpenHandle(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes, int iClose);
//	char cDebugFile[256];
	DWORD dwGot;
//	FILE *fpp;

	iMyThread = (int)InterlockedIncrement(&i_IO_Thread) - 1;

//	sprintf(cDebugFile, "F:\\IO_Thread_%d.txt", iMyThread);
//	fpp = fopen(cDebugFile, "w");
//	fprintf(fpp, "Thread %d starting\n", iMyThread);
//	fclose(fpp);

	while((int)i_IO_Thread < iN_IO_Threads) { _sleep(10); }
	if(iMyThread == 0) { iIO_Thread_Is_Up = 1; }
	while(iIO_Thread_Is_Up != 1) { _sleep(10); }

//	fpp = fopen(cDebugFile, "a+");
//	fprintf(fpp, "Thread %d started\n", iMyThread);
//	fclose(fpp);

	// Invalidate all cache buffers
	iCache = 0;
	lBufferOffset = (long)0;
	iWrap = 0;
	if(iMyThread == 0) {
		for(i=0; i<8192; i++) {
			iTile_of_iCache[i] = -1;
			iCache_of_iTile[i] = -1;
			iUsed_of_iCache[i] =  0;
		}
	}

	// Allocate memory
	lUnit = (long)64 * (long)1024;
	SIZE_T  dwSize   = (SIZE_T)iNCacheMB * (SIZE_T)1024 * (SIZE_T)1024;

	BobCache = (unsigned char *)VirtualAlloc(NULL, dwSize, MEM_COMMIT, PAGE_READWRITE);

	while(iIO_Thread_Is_Up == 1) {

		// Wait here until there is something to do
		while(iIO_Thread_Is_Up == 1  &&  iIO_Read == 0) { _sleep(50); }
		iPrinted = 0;

		// If iTile_of_iCache[0] does not point to any tile, then either the buffer cache has
		// been reset elsewhere or the cache wrapped on the previous pass.
		// This means we are starting from scratch: reset local counters and all cache pointers
		if(iTile_of_iCache[0] < 0) {
			iCache = 0;
			lBufferOffset = (long)0;
			iWrap = 0;
			if(iMyThread == 0) {
				for(i=0; i<8192; i++) {
					iTile_of_iCache[i] = -1;
					iCache_of_iTile[i] = -1;
					iUsed_of_iCache[i] = 0;
				} 
			}
		}

		// Count how much total buffer memroy will be needed in this pass assuming while keeping previous buffers
		lNeedTotalBufferSize = lBufferOffset;
		for(iRequest = 0; iRequest < iNRequests; iRequest++) {
			isHere = 0;
			iTile = iTile_of_iReq[iRequest];
			iCache0 = iCache_of_iTile[iTile];
			if(0 <= iCache0  &&  iCache0 < iCache) { if(iTile == iTile_of_iCache[iCache0]) { isHere = 1; }}
			if(isHere == 0) { lNeedTotalBufferSize += lUnit*(1 + ((long)dwSize_of_iReq[iRequest]-1)/lUnit); }
		}

		// If previous + new requests will run out of cache on this trip, wipe it all clean.
		// Otherwise, mark all cache buffers as used
		if(lNeedTotalBufferSize > (long)dwSize  ||  iWrap == 1) {
			iCache = 0;
			lBufferOffset = 0;
			iWrap = 0;
			if(iMyThread == 0) {
				for(i=0; i<8192; i++) {
					iTile_of_iCache[i] = -1;
					iCache_of_iTile[i] = -1;
					iUsed_of_iCache[i] = 0;
				} 
			}
		} else {
			if(iMyThread == 0) { for(i=0; i < iCache; i++) { iUsed_of_iCache[i] = 1; } }
		}

		// Sort the IO requests in blocks of iNSort;
		// Will assume tile order is consecutive in hv file
		if(iMyThread == 0) {
			iNSort =8;
			for(iChunk=0; iChunk<iNRequests; iChunk += iNSort) {
				iMax = iChunk+iNSort;
				if(iMax > iNRequests) { iMax = iNRequests; }
				for(i=iChunk; i<iMax; i++) { iTileTmp[i-iChunk] = iTile_of_iReq[i]; }
				for(i=iChunk; i<iMax; i++) {
					iReqMin = iChunk;
					for(ij=iChunk; ij<iMax; ij++) { if(iTileTmp[ij-iChunk] < iTileTmp[iReqMin-iChunk]) { iReqMin = ij; }}
					iSorted_Req_of_iReq[i] = iReqMin;
					iTileTmp[iReqMin-iChunk] = 9000;
				}
			}

//			sprintf(cDebugFile, "L:\\dhp\\projects\\h13\\IO_SORT.txt");
//			fpp = fopen(cDebugFile, "w");
//			fprintf(fpp, "start\n");
//			fprintf(fpp, "iNRequests = %d\n", iNRequests);
//			for(i=0; i<iNRequests; i++) {
//				fprintf(fpp, "i16=%2d   iReq=%3d   iSortReq=%3d   iTile=%3d\n",
//					i%16, i, iSorted_Req_of_iReq[i], iTile_of_iReq[iSorted_Req_of_iReq[i]]);
//			}
//			fprintf(fpp, "---------------------------------------------------------------\n");
//			fclose(fpp);
		}

		// Tell everyone that cache status is ready
		if(iMyThread == 0) {
			for(i=0; i<iN_IO_Threads; i++) { iThreadIsOnCache[i] = 0; }
			i_IO_Thread = 0;
			iCacheStatusIsReady = 1;
		} else {
			while(iCacheStatusIsReady != 1  &&  iIO_Read == 1) { _sleep(10); }
		}

		// Read in data
		iNConsecutive = 4;
		jRequest = 0;
		while(iIO_Read == 1  &&  jRequest < iNRequests) {
			iRequest = iSorted_Req_of_iReq[jRequest];

			// First, make sure that no IO thread is to far ahead of the others
			if(iN_IO_Threads > 1) {
				iThreadIsOnCache[iMyThread] = iCache;
				iCacheMin = -1000;
				while(iCacheMin < iCache-2*iNConsecutive*iN_IO_Threads  &&  iIO_Read == 1) {
					iCacheMin = iThreadIsOnCache[0];
					for(i=1; i<iN_IO_Threads; i++) {
						if(iCacheMin > iThreadIsOnCache[i]) { iCacheMin = iThreadIsOnCache[i]; }
					}
					if(iCacheMin < iCache-2*iNConsecutive*iN_IO_Threads) { _sleep(10); }
				}
			}

			// Determine if data is already cached
			isHere = 0;
			iTile = iTile_of_iReq[iRequest];
			iCache0 = iCache_of_iTile[iTile];
			if(0 <= iCache0  &&  iCache0 < iCache) { if(iTile == iTile_of_iCache[iCache0]) { isHere = 1; }}

			// Only read in tile if it is not already in cache
			if(isHere == 0) {
				// Generate range of buffer memory [lBufferOffset, lNextBufferOffset) for next read
				lNextBufferOffset = lBufferOffset + lUnit*(1 + ((long)dwSize_of_iReq[iRequest]-1)/lUnit);
				if(lNextBufferOffset > (long)dwSize) {
					lBufferOffset = (long)0;
					lNextBufferOffset = lUnit*(1 + ((long)dwSize_of_iReq[iRequest]-1)/lUnit);
					iWrap = 1;
				}
				// Wait for this range of memory to be free
				isFree = 0;
				while(isFree == 0  &&  iIO_Read == 1) {
					isFree = 1;
					for(i=0; i<iCache; i++) {
						if(iUsed_of_iCache[i] == 1) {
							iOverlap = 1;
							if(lNextBufferOffset <  lBufferMin_iCache[i]) { iOverlap = 0; }
							if(    lBufferOffset >= lBufferMax_iCache[i]) { iOverlap = 0; }
							if(iOverlap == 1) { isFree = 0; }
						}
					}
					if(isFree == 0) { _sleep(10); }
				}

				if(iIO_Read == 1) {

					if((jRequest/iNConsecutive)%iN_IO_Threads == iMyThread) {

						// Invalidate the cache buffers which previously used this memory
// -----------------------------------------------------------------------------------------------
// I dont think that this is necessary
//						for(i=0; i<iCache; i++) {
//							iOverlap = 1;
//							if(lNextBufferOffset <  lBufferMin_iCache[i]) { iOverlap = 0; }
//							if(    lBufferOffset >= lBufferMax_iCache[i]) { iOverlap = 0; }
//							if(iOverlap == 1) {
//								iTile0 = iTile_of_iCache[i];
//								iTile_of_iCache[i] = -1;
//								if(iTile0 > 0) { iCache_of_iTile[iTile0] = -1; }
//							}
//						}
// -----------------------------------------------------------------------------------------------

						// Read data into this cache buffer
						lpBuffer_of_iCache[iCache] = &BobCache[lBufferOffset];
						sprintf(cFileName, "%s", cFileName_of_iReq[iRequest]);
						cFileName[0] += iMyThread;

//						fpp = fopen(cDebugFile, "a+");
//						fprintf(fpp, "Thread %d reading %s  iReq=%3d  iCache=%3d  offset=%ld\n",
//							iMyThread, cFileName, iRequest, iCache, lBufferOffset);
//						fclose(fpp);

						if(m_iBufferedIO != 1) {
							dwGot = dhpReadFromFileSync(lpBuffer_of_iCache[iCache],
														cFileName,
														i64FileOffset_of_iReq[iRequest],
														dwSize_of_iReq[iRequest]);
						} else {
							dwGot = dhpReadFromFileNoBuf(lpBuffer_of_iCache[iCache],
														 cFileName,
														 i64FileOffset_of_iReq[iRequest],
														 dwSize_of_iReq[iRequest]);
						}

						// Validate new cache buffers in this location
						iTile_of_iCache[iCache] = iTile;
						iCache_of_iTile[iTile] = iCache;
						lBufferMin_iCache[iCache] = lBufferOffset;
						lBufferMax_iCache[iCache] = lNextBufferOffset;
						iUsed_of_iCache[iCache] = 1;
					}       // End of (iRequest/iNConsecutive)%iN_IO_Threads == iMyThread)  IF-BLOCK
					lBufferOffset = lNextBufferOffset;
					iCache++;
				}
			}				// End of (isHere == 0) IF-BLOCK
			jRequest++;
		}					// End of iNRequest while loop

		// Wait for all of the IO threads to finish before moving on
		i = (int)InterlockedIncrement(&i_IO_Thread);
		while(i_IO_Thread < iN_IO_Threads) { _sleep(10); }

		if(iMyThread == 0) {
			iIO_Read = 0;
			iNRequests = 0;
		} else {
			while(iIO_Read != 0) { _sleep(10); }
		}
	}						// End of (iIO_Thread_Is_Up == 1) WHILE-LOOP

	// Tell everyone that IO thread has exitted           iCache
	if(iMyThread == 0) {
		iIO_Thread_Is_Up = 0;
		i_IO_Thread = 0;
	}
}

void Chvr_work::SpawnWork()
{

    int depth = 0, width, height;
	BYTE  colortype;
	DWORD buffertype;
	char s[4096];
	HWND CreateOpenGLWindow(char* title, int x, int y, int width, int height, 
		   BYTE type, DWORD flags, int depth, HWND hLocalWnd);
	void deleteGLPallet();
    void  setstr(char*, char*);


/************************************************************
 *  Here's how to create an OpenGL window using wgl         *
 ************************************************************/
	*s = '\0';
	strcat(s,"HVR - SERVER");

	colortype = PFD_TYPE_RGBA;
//	colortype = PFD_TYPE_COLORINDEX;

	buffertype = PFD_DOUBLEBUFFER;
//	buffertype = 0;
	depth = 32;

	if(iGLwindowIsMade == 0) {

		width = m_ShvrIn.iWidth;
		height = m_ShvrIn.iHeight;

		hWnd = CreateOpenGLWindow(s, 0, 0, width, height, colortype, buffertype, depth, m_hLocalWnd);
		if (hWnd == NULL) { exit(1); }
		hDC = GetDC(hWnd);
		hRC = wglCreateContext(hDC);
		wglMakeCurrent(hDC, hRC);

		ShowWindow(hWnd, SW_SHOW);
		UpdateWindow(hWnd);

		init();

		iGLwindowIsMade = 1;
	} else {
		wglMakeCurrent(hDC, NULL);
		wglDeleteContext(hRC);
		if(m_hLocalWnd == NULL) { DestroyWindow(hWnd); }
		deleteGLPallet();

		width = m_ShvrIn.iWidth;
		height = m_ShvrIn.iHeight;
		hWnd = CreateOpenGLWindow(s, 0, 0, width, height, colortype, buffertype, depth, m_hLocalWnd);

		hDC = GetDC(hWnd);
		hRC = wglCreateContext(hDC);
		wglMakeCurrent(hDC, hRC);

		ShowWindow(hWnd, SW_SHOW);
		UpdateWindow(hWnd);

		init();

		isubcur = -1;         /* current sub-brick loaded into bobsub */
		sprintf(m_cOldHVname, "none");
	}

	// If hvr_server is being driven remotely, turn cursor off
	if(m_hLocalWnd == NULL) { ShowCursor(FALSE); }

	while(iGLwindowIsMade == 1) {
		if(iDoThis == 0) {
			_sleep(10);
		} else {
			if(iDoThis == 1) {
				draw_scene();
				iStopRendering = 0;
				iDoThis = 0;
			}
			if(iDoThis == 2) {
				// Shut down OpenGL environment IF last client
				iNClients--;
				if(iNClients == 0) {
					void deleteGLPallet();

					wglMakeCurrent(hDC, NULL);
					wglDeleteContext(hRC);
					if(m_hLocalWnd == NULL) { DestroyWindow(hWnd); }
					deleteGLPallet();

					isubcur = -1;         /* current sub-brick loaded into bobsub */
					sprintf(m_cOldHVname, "none");
					iGLwindowIsMade = 0;
				}
				iDoThis = 0;
			}
			if(iDoThis == 3) {
				DoSetOption();
				iDoThis = 0;
			}
			if(iDoThis == 4) {
				DoPutPix();
				iDoThis = 0;
			}
			if(iDoThis == 5) {
				DoGetPix();
				iDoThis = 0;
			}
		}

//		if(my_work->DoRender()) { my_work->draw_scene(); }
//		else                    { SleepEx(10, FALSE); }

	}
}

STDMETHODIMP Chvr_work::StartUp()
{
	// Registar that there is one more client for HVR server
	iNClients++;

	// If redering thread is already up then no need to start rendering or IO threads, so return
	if(iGLwindowIsMade == 1) { return S_OK; }

	// Start a rendering thread
	_beginthread(OpenAWindow, 0, (void *) this);
/***************************************************************************
 * Here is how you start a rendering thread with GLUT,
 * Note, all of the acctions in SpawnWork would need to be GLUT callbacks
 ***************************************************************************
	int argc;
	char *argv[100];
    pParam = this;
    CopyHvrInputs();
	argc = 2;
	argv[0] = "hvr";
	argv[1] = "tri";
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_SINGLE);
	glutCreateWindow("hvr GOOP GLUT");
	glutReshapeWindow(m_ShvrUse.iWidth, m_ShvrUse.iHeight);
	glutDisplayFunc(OpenAWindow);
	glutMainLoop();
***************************************************************************/

	// Start IO threads
	int i;
	for(i=0; i<iN_IO_Threads; i++) { _beginthread(StartIOThread, 0, (void *) this); }

	return S_OK;
}

void Chvr_work::draw_scene()
{
    float eye[3], up[3], torq[3], eps, eps1;
    int i, j, joff, k, ic, ib, jb, kb, ijk, nslices, ntiles, ididread;
    int isub, jsub, ixoff, iyoff, izoff, itick, jlutmax;
	int i1, i2;
    int nxsub, nysub, nzsub, nr, iprint, iTexBndry;
    int nxsb2, nysb2, nzsb2;
    int isoutside_n, isoutside_f, isoutside_t;
    int isoutside_b, isoutside_l, isoutside_r;
//	int fRBits, fGBits, fBBits, fABits;
//	float alphafac;

    float facXtoN;
    float subvolsiz[3], subminpos[3], fulminpos[3], submaxpos[3];
    float alpha, alpha0;
    float xx, yy, zz, dxx, dyy, dzz, axx, ayy, azz, dz, dx, dy;
	int   idir, ids0, idsd;
    int isBehind;
    float vhh, vnn, vvv, vhhmin, vhhmax, vvvmin, vvvmax, vhhcent, vvvcent;
    float vnormi;
	float xxv, yyv, zzv;
//    float dsmax, dsmin, sfinish,ss, dsfar, dsnear, fac;
    float edgedist, edgemin, edgemax;
    float hdistfac,vdistfac;
    float eyepos[3],cubesz[3];
	static char old_amap[1024] = "default";
//    static char old_hvname[1024] = "none";
	DWORD dwGot;
	char str[256], str_tmp[256];

    GLdouble dleft,dright,dbottom,dtop, dnearclip,dfarclip;

/* frustum stuff */
    void crossprod(float vec1[3], float vec2[3], float vec3[3]);
/* frustum stuff */

/* AuxClip stuff */
	int iClipThis, iClipThis0;
	float px0, px1, py0, py1, pz0, pz1, px0c, px1c, py0c, py1c, pz0c, pz1c;
	float xx0c, xx1c, yy0c, yy1c, zz0c, zz1c, zz0, zz1;
/* AuxClip stuff */

/* subtile stuff */
    float fminx,fminy,fminz,distmax, tile_dist[2048];
    int ippp, iccc, nccc, nlist, list_tiles[2048], tile_did[2048];
	int iNTileRender, iTileRender[2048];
    int inode, ilist, nlistnew, need_check, list_new[2048];
	int ixx, iyy, izz, jparent, nrj;
    float delxmin, delx, xmin, xmax, ymin, ymax, zmin, zmax;
    float distmin, delxtol, delxsize;
/* subtile stuff */

/* subtile io stuff */
      static char *seps = " ,;\t\n";
      static int nsub, node0;
      static long tile_off64t[2048];
      static int tile_nrep[2048];
      static int tile_nxsub[2048], tile_nysub[2048], tile_nzsub[2048];
      static int tile_ixoff[2048], tile_iyoff[2048], tile_izoff[2048];
      static int tile_parent[2048];
      static int tile_maxerr[2048];
      static int ncPath;
      static char tile_file[2048][64], tile_path[64], tile_full[256];
      static char line[256];
      static int tile_children[2048][32];
	  static float fLimits[6];
//	  FILE *fplim;
	  FILE *fpp;
/* subtile io stuff */

/* IO stuff */
    int nbytes;
//	int nbin;
//	FILE *fdd;
    long offset64t;
	int iOpac, iRGB, iALP;
	char tile_fileO[256];
	unsigned char *bobin;
/* IO stuff */

/* multiple pannel stuff */
    int   ip;
	int pan_idir[NPMAX];
    float d, dhorz, dvert, pcn, pan_cenmagi[NPMAX];
    float pan_eye[NPMAX][3],pan_up[NPMAX][3],pan_center[NPMAX][3];
    float pan_lookat[NPMAX][3],pan_torq[NPMAX][3],pan_dlef[NPMAX];
    float pan_drit[NPMAX],pan_dbot[NPMAX],pan_dtop[NPMAX];
    float pan_lefpar[NPMAX][3],pan_ritpar[NPMAX][3],pan_toppar[NPMAX][3];
    float pan_botpar[NPMAX][3],pan_lef_o[NPMAX][3],pan_rit_o[NPMAX][3];
    float pan_top_o[NPMAX][3], pan_bot_o[NPMAX][3];
/* multiple pannel stuff */

/* slice stuff */
	int islice0, islice1, isliceoff0, isliceoff1;
	float dx1, dy1, dz1, ppv0, ppv1, wt0;
	unsigned long lrgba0[256], lrgba1[256];
    int islicesize, islice, isliceoff;
    float xx0, xx1, yy0, yy1;
	float fZero, fOne, fMOne, xxv0, xxv1, yyv0, yyv1;
//	GLfloat fCurRastPos[4];
//	float fSMid, fLowLef[3], fLowRit[3], fTopLef[3], fTopRit[3];
/* slice stuff */

/* Statistics stuff (stats) */
	int i0, iPDF[256], iCross[256*256];
	long lBlack, lWhite, lGreen, lCrossLUT[256];
	int   ind[10] = {    0,     1,    33,    65,    97,   129,   161,   193,   225, 256  };
	double red[10] = {255.9,   0.0,   0.0,   0.0, 160.0, 160.0,   0.0,   0.0,   0.0, 255.9};
	double grn[10] = {255.9,   0.0, 160.0,   0.0,   0.0, 160.0, 160.0, 255.9,   0.0,   0.0};
	double blu[10] = {255.9,   0.0,   0.0, 160.0,   0.0,   0.0, 160.0,   0.0, 255.9,   0.0};
	float fRed, fGrn, fBlu;
	int jMax, iMaxCounts, iTotalCounts;
	double fDelX, fPDFfactor, f10PDFmax;
	float xPDFmin, xPDFmax;
	int i10PDFmax;
//HFONT hfnt, hfntPrev; 
//PLOGFONT plf;
//RECT rc;
/* Statistics stuff (stats) */

/* prototyping functions */
    int match( char *token, char *word, int tokenlength );
    float dotproduct( const float*, const float* );
	void vnormalize( float v[3] );
    int   strdif(char*, char*);
    void  setstr(char*, char*);
    int readCmap( char *clrname, char *alpname, float ftable[256][4] );
	void TposeYZ( unsigned char *a1, unsigned char *a2, int *nx2, int *ny2, int *nz2,
														int *nx,  int *ny,  int *nz  );
	void TposeXZ( unsigned char *a1, unsigned char *a2, int nx2, int ny2, int nz2,
														int nx,  int ny,  int nz  );
	DWORD dhpReadFromFileNoBuf(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes);
	DWORD  dhpReadFromFileSync(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes);
	DWORD  dhpReadFromOpenFile(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes,
		int iFile, int iClose);
/* prototyping functions */

	SIZE_T dwNeedSize, dwMaxNeededSize;

/*
	FILE *fp;
	GLint params;
	glGetIntegerv(GL_ACCUM_GREEN_BITS, &params);
	fp = fopen("C:\\users\\dhp\\hvrDCOM\\data\\bitplanes.txt", "w");
	fprintf(fp, "Number of Green bit planes = %d\n", params);
	fclose(fp);
*/
	// Copy from GUI input struct to hvr use struct
	// Set all iNew flags to the status of NOT new.
    CopyHvrInputs();

	// IF a gl color table is available, then  sure color table is set
	if(m_bColorTableIsAvailable) {
		glColorTableEXT(DHP_TEX, GL_RGBA,256,GL_RGBA, GL_FLOAT,frgba);
		glFlush();
		glFinish();
	}

	if(m_iTimersOn == 1) {
		tsEvent.SetOutputFileName(m_sTimerFile);
		tsEvent.StampTime("Begin Rendering");
	}

    iprint = 0;

	// Force re-read of all data if doing cross-correlation
	if(strcmp(m_sStats, "cross") == 0) { sprintf(m_cOldHVname, "none"); }

	dwMaxNeededSize = 0;
    if(strdif(m_cOldHVname, m_ShvrUse.sHVname) == 1) {
	  FlushCacheBuffers();
      setstr(m_cOldHVname, m_ShvrUse.sHVname);
      fpp = fopen(m_ShvrUse.sHVname, "r");
	  if(fpp == NULL) {
		tile_nrep[0] = 1;
		tile_nxsub[0]  = 64;
		tile_nysub[0]  = 64;
		tile_nzsub[0]  = 64;
		tile_ixoff[0]  =  0;
		tile_iyoff[0]  =  0;
		tile_izoff[0]  =  0;
		tile_parent[0] = -1;
		tile_maxerr[0] =  0;
		m_iNBoundaries = 0;
		for(i=0; i<128; i++) { m_iNBoundariesRep[i] = 0; }
		nsub = 1;
		node0 = 0;
		isubcur = 0;
		int ith = 2;
		bobsubZ = bobsubZ0;
		for(k=0; k<tile_nzsub[0]; k++) { kb=0; if(k<ith || k>tile_nzsub[0]-ith-1) { kb = 1; }
		for(j=0; j<tile_nysub[0]; j++) { jb=0; if(j<ith || j>tile_nysub[0]-ith-1) { jb = 1; }
		for(i=0; i<tile_nxsub[0]; i++) { ib=0; if(i<ith || i>tile_nxsub[0]-ith-1) { ib = 1; }
			ijk = i + tile_nxsub[0] * (j + tile_nysub[0] * k);
			if(kb+jb+ib > 1) { 
				bobsubZ[ijk] = (unsigned char)(127 + 
					(int)((float)128 * ((float)i/(float)tile_nxsub[0] +
					                    (float)j/(float)tile_nysub[0] +
										(float)k/(float)tile_nzsub[0]   ) / (float)3));
			} else {
				bobsubZ[ijk] = (unsigned char)0;
			}
		}}}
		for(ijk=0; ijk<tile_nxsub[0]*tile_nysub[0]*tile_nzsub[0]; ijk++) {
			bobsubY[ijk] = bobsubZ[ijk];
		}
		bSetBobX = FALSE;
		bSetBobY = FALSE;

		fLimits[0] = -(float)1.0;
		fLimits[1] =  (float)1.0;
		fLimits[2] = -(float)1.0;
		fLimits[3] =  (float)1.0;
		fLimits[4] = -(float)1.0;
		fLimits[5] =  (float)1.0;
	  } else {
		for(i=0; i<128; i++) { m_iNBoundariesRep[i] = 0; }
		nsub = 0;
		node0 = -1;
		while (node0 < 0) {
			i = match(fgets(line,200,fpp), "end", 3);
			strncpy(tile_file[nsub], strtok( line,seps ), 64);
			tile_off64t[nsub] = (long)atoi( strtok(NULL,seps) );
			tile_nrep[nsub]   = atoi( strtok(NULL,seps) );
			tile_nxsub[nsub]  = atoi( strtok(NULL,seps) );
			tile_nysub[nsub]  = atoi( strtok(NULL,seps) );
			tile_nzsub[nsub]  = atoi( strtok(NULL,seps) );
			tile_ixoff[nsub]  = atoi( strtok(NULL,seps) );
			tile_iyoff[nsub]  = atoi( strtok(NULL,seps) );
			tile_izoff[nsub]  = atoi( strtok(NULL,seps) );
			tile_parent[nsub] = atoi( strtok(NULL,seps) );
			tile_maxerr[nsub] = atoi( strtok(NULL,seps) );
			if(tile_parent[nsub] < 0) { node0 = nsub; }

			if( m_iNBoundariesRep[tile_nrep[nsub]] < -tile_ixoff[nsub] ) {
				m_iNBoundariesRep[tile_nrep[nsub]] = -tile_ixoff[nsub];
			}

			// Get Max memory needy for any brick in this hv file
			// This wiil get used below to determine if bob arrays need to be resized
			nxsub = tile_nxsub[nsub];
			nysub = tile_nysub[nsub];
			nzsub = tile_nzsub[nsub];
			nxsb2 = 2;   while(nxsb2 < nxsub) { nxsb2 *= 2; }
			nysb2 = 2;   while(nysb2 < nysub) { nysb2 *= 2; }
			nzsb2 = 2;   while(nzsb2 < nzsub) { nzsb2 *= 2; }
			dwNeedSize = (SIZE_T)(nxsb2 * nysb2 * nzsb2);
			if(dwMaxNeededSize < dwNeedSize) { dwMaxNeededSize = dwNeedSize; }

			nsub++;
		}

		// Set default limits
		/* modelview aspect for the entire mesh is based on X size. */
		m_iNBoundaries = m_iNBoundariesRep[tile_nrep[node0]];
		Nx = tile_nrep[node0] * (tile_nxsub[node0] - 2*m_iNBoundaries);  // These will need to be generalized
		Ny = tile_nrep[node0] * (tile_nysub[node0] - 2*m_iNBoundaries);  // to be based on highest res tiles.
		Nz = tile_nrep[node0] * (tile_nzsub[node0] - 2*m_iNBoundaries);

		fLimits[0] = -(float)1.0;
		fLimits[1] =  (float)1.0;
		fLimits[2] = -(float)Ny/(float)Nx;
		fLimits[3] =  (float)Ny/(float)Nx;
		fLimits[4] = -(float)Nz/(float)Nx;
		fLimits[5] =  (float)Nz/(float)Nx;

		// Set limits to those specified in hv file
		if(match(fgets(line,200,fpp), "XYZ_Limits", 10) == 1) {
			strncpy(str_tmp, strtok( line,seps ), 64);
			fLimits[0] = (float)atof( strtok(NULL,seps) );
			fLimits[1] = (float)atof( strtok(NULL,seps) );
			fLimits[2] = (float)atof( strtok(NULL,seps) );
			fLimits[3] = (float)atof( strtok(NULL,seps) );
			fLimits[4] = (float)atof( strtok(NULL,seps) );
			fLimits[5] = (float)atof( strtok(NULL,seps) );
//			fplim = fopen("hvr_limits.txt", "w");
//			fprintf(fplim, "Limits X: %f %f\n", fLimits[0], fLimits[1]);
//			fprintf(fplim, "Limits Y: %f %f\n", fLimits[2], fLimits[3]);
//			fprintf(fplim, "Limits Z: %f %f\n", fLimits[4], fLimits[5]);
//			fclose(fplim);
		}
		fclose(fpp);
		isubcur = -1;         /* current sub-brick loaded into bobsub */

		// Check and fix tree structure if needed
		for(i=0; i<nsub; i++) { if(i != node0) {
			nr = tile_nrep[i];
			ixx = nr*(tile_ixoff[i]+tile_nxsub[i]/2);
			iyy = nr*(tile_iyoff[i]+tile_nysub[i]/2);
			izz = nr*(tile_izoff[i]+tile_nzsub[i]/2);
			jparent = -1;
			for(j=0; j<nsub; j++) { if(tile_nrep[j] == 2*nr) {
				nrj = tile_nrep[j];
				if((nrj*tile_ixoff[j]-ixx)*(nrj*(tile_ixoff[j]+tile_nxsub[j])-ixx) < 0
				&& (nrj*tile_iyoff[j]-iyy)*(nrj*(tile_iyoff[j]+tile_nysub[j])-iyy) < 0
				&& (nrj*tile_izoff[j]-izz)*(nrj*(tile_izoff[j]+tile_nzsub[j])-izz) < 0) {
					jparent = j;
				}
			}}
			if(jparent < 0) {
				fpp = fopen("HVR_TreeError.out", "w");
				fprintf(fpp, "File = %s\n", m_ShvrUse.sHVname);
				fprintf(fpp, "no parent for node %i\n", i);
				fprintf(fpp, "nr=%d  i[xyz]off= %d %d %d\n",
					nr,tile_ixoff[i],tile_iyoff[i],tile_izoff[i] );
				fclose(fpp);
				return;
			}
			if(jparent != tile_parent[i]) {
				fpp = fopen("HVR_TreeFix.out", "a+");
				fprintf(fpp, "File = %s\n", m_ShvrUse.sHVname);
				fprintf(fpp, "Incorrect parent for node %i\n", i);
				fprintf(fpp, "nr=%d  i[xyz]off= %d %d %d\n",
					nr,tile_ixoff[i],tile_iyoff[i],tile_izoff[i] );

				fprintf(fpp, "Incorrect parent in file is:\n");
				j = tile_parent[i];
				fprintf(fpp, "node=%d   nr=%d  i[xyz]off= %d %d %d\n",
					j, nr,tile_ixoff[j],tile_iyoff[j],tile_izoff[j] );

				fprintf(fpp, "Correct parent according to subdomains is:\n");
				j = jparent;
				fprintf(fpp, "node=%d   nr=%d  i[xyz]off= %d %d %d\n",
					j, nr,tile_ixoff[j],tile_iyoff[j],tile_izoff[j] );

				fclose(fpp);
				tile_parent[i] = jparent;
			}
		}}
	  }
    }

	// Resize bob arrays if needed
	if(dwMaxSize < dwMaxNeededSize) {
		VirtualFree(bobsubX  , dwMaxSize, MEM_DECOMMIT | MEM_RELEASE);
		VirtualFree(bobsubY  , dwMaxSize, MEM_DECOMMIT | MEM_RELEASE);
		VirtualFree(bobsubZ0 , dwMaxSize, MEM_DECOMMIT | MEM_RELEASE);
		VirtualFree(bobOX    , dwMaxSize, MEM_DECOMMIT | MEM_RELEASE);
		VirtualFree(bobOY    , dwMaxSize, MEM_DECOMMIT | MEM_RELEASE);
		VirtualFree(bobOZ    , dwMaxSize, MEM_DECOMMIT | MEM_RELEASE);

		dwMaxSize = dwMaxNeededSize;

		bobsubX  = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
		bobsubY  = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
		bobsubZ0 = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
		bobOX    = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
		bobOY    = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
		bobOZ    = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
		bBobsAllocated = TRUE;
	}

	// Test for an independent opacity field: switch is RGB in name of hv file
	// 01234567890
	// v72RGB_wa0123.hv
	if(	tile_file[0][3] == 'R' &&
		tile_file[0][4] == 'G' &&
		tile_file[0][5] == 'B'		) { iOpac = 1; } else { iOpac = 0; }

    ncPath = 0;
    for(i=0; i<100; i++) {
		if(m_ShvrUse.sHVname[i] ==  '/') { ncPath = i+1; }
		if(m_ShvrUse.sHVname[i] == '\\') { ncPath = i+1; }
	}
    for(i=0; i<ncPath; i++) { tile_path[i] = m_ShvrUse.sHVname[i]; }
    tile_path[ncPath] = '\0';

    nbytes = m_ShvrUse.iHeight * m_ShvrUse.iWidth * 3;

    if(pannels_initted == 0) {
      for(ip=0; ip<npannel; ip++) { rgbpannel[ip]  = (char*)malloc(nbytes); }
      pannels_initted = 1;
    }

    for(ip=0; ip<npannel; ip++) {
      for(k=0; k<nbytes; k++) { rgbpannel[ip][k] = 0; }
    }
    ipcur = -1;

   /************************************************************************
    *  Generate horizontal, vertical, and alpha scalings                   *
    ************************************************************************/
    vdistfac = (float)tan(0.5*Pid180 * m_ShvrUse.fFovAngle);
	hdistfac = vdistfac / m_ShvrUse.fAspectRatio;
    alpha0 = opacity * (float)128.0;

    dnearclip = m_ShvrUse.fNearClip;
    dfarclip  = m_ShvrUse.fFarrClip;
//    dx = hdistfac * dnearclip;
//    dy = vdistfac * dnearclip;

   /************************************************************************
    * Right handed ONB  which power-wall frame is attached to :            *
    * (eye[], up[], torq[])                                                *
    ************************************************************************/
    for(i=0; i<3; i++) {
		eye[i]    = m_ShvrUse.fCent[i] - m_ShvrUse.fEye[i];	/* direction eye looks in */
		up[i]     = m_ShvrUse.fUp[i];
		eyepos[i] = m_ShvrUse.fEye[i];
	}
    vnormalize(eye);

    /* Insist that up is normal to eye */
    d = dotproduct(eye,up);
    for (i=0; i<3; ++i) { up[i] = up[i]- eye[i]*d; }
    vnormalize( up );

    /* Construct the cross vector 'torque' */
    torq[0] = eye[1]*up[2] - eye[2]*up[1];
    torq[1] = eye[2]*up[0] - eye[0]*up[2];
    torq[2] = eye[0]*up[1] - eye[1]*up[0];
/*
	fpp = fopen("rend_ONB.out", "w");
	fprintf(fpp, "NEW ONE\n");
    fprintf(fpp, "eye  = %f %f %f\n", eye[0], eye[1], eye[2]);
    fprintf(fpp, "up   = %f %f %f\n", up[0],  up[1],  up[2]);
    fprintf(fpp, "torq = %f %f %f\n", torq[0],torq[1],torq[2]);
      for(ip=0; ip<npannel; ip++) {
		fprintf(fpp, "ip = %d\n", ip);
        fprintf(fpp, "pw_normal= %f %f %f\n",
                pw_normal[ip][0],pw_normal[ip][1],pw_normal[ip][2]);
        fprintf(fpp, "pw_up    = %f %f %f\n",
                pw_up[ip][0],pw_up[ip][1],pw_up[ip][2]);
        fprintf(fpp, "pw_center= %f %f %f\n",
                pw_center[ip][0],pw_center[ip][1],pw_center[ip][2]);
      }

	fclose(fpp);
*/
    for(ip=0; ip<npannel; ip++) {
      for(i=0; i<3; i++) {
        pan_eye[ip][i]    = m_ShvrUse.fPanNorm[0] * torq[i] + 
                            m_ShvrUse.fPanNorm[1] *  eye[i] +
                            m_ShvrUse.fPanNorm[2] *   up[i];

        pan_up[ip][i]     = m_ShvrUse.fPanUp[0] * torq[i] + 
                            m_ShvrUse.fPanUp[1] *  eye[i] +
                            m_ShvrUse.fPanUp[2] *   up[i];

        pan_center[ip][i] = m_ShvrUse.fPanCent[0] * torq[i] + 
                            m_ShvrUse.fPanCent[1] *  eye[i] +
                            m_ShvrUse.fPanCent[2] *   up[i];

        pan_lookat[ip][i] = eyepos[i] + pan_eye[ip][i];
      }
/* orientation dependent only on eye
	  dxx = eye[0];
	  dyy = eye[1];
	  dzz = eye[2];
*/

	  // Normalize panel eye, up, and torq vectors
      vnormalize(pan_eye[ip]);
      vnormalize(pan_up[ip]);
      crossprod(pan_eye[ip], pan_up[ip], pan_torq[ip]);
      vnormalize(pan_torq[ip]);

      // Default orientation of rendering planes: used if to much of the subcube is behind you */
	  dxx = pan_eye[ip][0];
	  dyy = pan_eye[ip][1];
	  dzz = pan_eye[ip][2];

	  if(dxx > 0.) { axx = dxx; } else { axx = -dxx; }
	  if(dyy > 0.) { ayy = dyy; } else { ayy = -dyy; }
	  if(dzz > 0.) { azz = dzz; } else { azz = -dzz; }
	  pan_idir[ip] = 2;                                     // view along z-direction
	  if(ayy > axx  &&  ayy > azz) { pan_idir[ip] = 1; }    // view along y-direction
	  if(axx > ayy  &&  axx > azz) { pan_idir[ip] = 0; }    // view along x-direction

     /***************************************************************
      * pan_center is position of the center (relative to the eye) of the pannel
      *   to be used in generating horizontal and vertical offset of
      *   pannel window
      * pan_lookat is a point in the direction normal to the pannel
      *   to be used in glLookAt
      * NOTE: the pannel must be centered on the normal to the eyeposition
      *       for pan_center & pan_lookat to be the same
      * NOTE: pan_center is NOT normalized
      *       ss * pan_dlef ... are shifts on a plane at a distance ss
      *       pan_cenmagi[ip] is the correction factor.
      ***************************************************************/
      pan_cenmagi[ip] = (float)1. / dotproduct(pan_eye[ip], pan_center[ip]);
      dhorz = dotproduct(pan_center[ip], pan_torq[ip]);
      dvert = dotproduct(pan_center[ip], pan_up[ip]);

      pcn = (float)1. / pan_cenmagi[ip];
      pan_dlef[ip] = (dhorz - m_ShvrUse.fHalfWidth ) * pan_cenmagi[ip];
      pan_drit[ip] = (dhorz + m_ShvrUse.fHalfWidth ) * pan_cenmagi[ip];
      pan_dbot[ip] = (dvert - m_ShvrUse.fHalfHeight) * pan_cenmagi[ip];
      pan_dtop[ip] = (dvert + m_ShvrUse.fHalfHeight) * pan_cenmagi[ip];
/*
	  fpp = fopen("rend_ONB.out", "a+");
	  fprintf(fpp, "ip = %d\n", ip);
	  fprintf(fpp, "Pid180= %f, fovangle=%f Aspect=%f\n", Pid180, m_ShvrUse.fFovAngle, m_ShvrUse.fAspectRatio);
	  fprintf(fpp, "pan_drit= %f, dhorz=%f\n", pan_drit[ip], dhorz);
	  fprintf(fpp, "half_width= %f, pan_cenmagi[ip]=%f\n", m_ShvrUse.fHalfWidth, pan_cenmagi[ip]);
	  fprintf(fpp, "pan_up     = %f %f %f\n",pan_up[ip][0],pan_up[ip][1],pan_up[ip][2]);
	  fprintf(fpp, "pan_torq   = %f %f %f\n",pan_torq[ip][0],pan_torq[ip][1],pan_torq[ip][2]);
	  fprintf(fpp, "pan_center = %f %f %f\n",pan_center[ip][0],pan_center[ip][1],pan_center[ip][2]);
	  fclose(fpp);
*/
    }

	/* modelview aspect for the entire mesh is based on X size. */
	m_iNBoundaries = m_iNBoundariesRep[tile_nrep[node0]];
	Nx = tile_nrep[node0] * (tile_nxsub[node0] - 2*m_iNBoundaries);  // These will need to be generalized
	Ny = tile_nrep[node0] * (tile_nysub[node0] - 2*m_iNBoundaries);  // to be based on highest res tiles.
	Nz = tile_nrep[node0] * (tile_nzsub[node0] - 2*m_iNBoundaries);

    cubesz[0]= (float)0.5 * (fLimits[1] - fLimits[0]);
    cubesz[1]= (float)0.5 * (fLimits[3] - fLimits[2]);
    cubesz[2]= (float)0.5 * (fLimits[5] - fLimits[4]);

    for(ip=0; ip<npannel; ip++) {
     /************************************************************************
      * Generate outward orthogonal surface vectors for 4 planes of the      *
      * viewing frustum.  They do not need to be normalized.                 *
      * These are used below to determine if a tile is OUTSIDE the frustum   *
      * Power wall change:  In principle, need to do these 4 planes for each *
      * pannel of a power wall.  In practice, it is actually easiest to, in  *
      * fact, do each of the pannels individuallys.  The condition on which  *
      * a tile is read in is simply that any part of the tile is in any one  *
      * of the pannels viewing frustum.                                      * 
      ************************************************************************/
      for(i=0; i<3; i++) {
        pan_lefpar[ip][i] = pan_eye[ip][i] + pan_dlef[ip]*pan_torq[ip][i];
        pan_ritpar[ip][i] = pan_eye[ip][i] + pan_drit[ip]*pan_torq[ip][i];
        pan_toppar[ip][i] = pan_eye[ip][i] + pan_dtop[ip]*pan_up[ip][i];
        pan_botpar[ip][i] = pan_eye[ip][i] + pan_dbot[ip]*pan_up[ip][i];
      }
                                            /* outward orthogonal vector at */
      crossprod(pan_up[ip],     pan_lefpar[ip], pan_lef_o[ip]); /*     left */
      crossprod(pan_ritpar[ip], pan_up[ip],     pan_rit_o[ip]); /*    right */ 
      crossprod(pan_torq[ip],   pan_toppar[ip], pan_top_o[ip]); /*      top */
      crossprod(pan_botpar[ip], pan_torq[ip],   pan_bot_o[ip]); /*   bottom */
/*
	  fpp = fopen("rend_ONB.out", "a+");
	  fprintf(fpp, "ip = %d\n", ip);
	  fprintf(fpp, "pan_ritpar= %f %f %f\n", pan_ritpar[ip][0],pan_ritpar[ip][1],pan_ritpar[ip][2]);
	  fprintf(fpp, "pan_up    = %f %f %f\n", pan_up[ip][0],pan_up[ip][1],pan_up[ip][2]);
	  fprintf(fpp, "pan_rit_o = %f %f %f\n", pan_rit_o[ip][0],pan_rit_o[ip][1],pan_rit_o[ip][2]);
	  fclose(fpp);
*/
    }

    if(iprint == -1) { exit(0); }

/***************************************************************************
 * INPUT: color map information                                            *
 ***************************************************************************/
    if(strdif(m_ShvrUse.sCmapname, m_cOldCmap) == 1) {
      setstr(m_cOldCmap, m_ShvrUse.sCmapname);
      if(strdif(m_ShvrUse.sCmapname, "default") == 1) {
		  ididread = readCmap( m_ShvrUse.sCmapname, m_ShvrUse.sAmapname, frgba );

		  if(ididread < 0) {
			  fpp = fopen("rend_cmap_err.txt", "w");
			  fprintf(fpp, "ididread = %d\n", ididread);
			  fprintf(fpp, "cmap: -->%s<--\n", m_ShvrUse.sCmapname);
			  fprintf(fpp, "amap: -->%s<--\n", m_ShvrUse.sAmapname);
			  for(i=0; i<256; i++) {
				  fprintf(fpp, "%3d %f %f %f %f\n", i, frgba[i][0], frgba[i][1], frgba[i][2], frgba[i][3]);
			  }
			  fclose(fpp);

//				MessageBox(hWnd, "Can NOT open alpha/color maps", "LALA", MB_OK);
		  }
		  if(ididread < 0) { return; }
	  }
	  glBindTexture(DHP_TEX, texname[1]);

	  if(m_bColorTableIsAvailable) { glColorTableEXT(DHP_TEX, GL_RGBA,256,GL_RGBA, GL_FLOAT,frgba); }

	  for(i=0; i<256; i++) {
		  m_lrgba[i] =						  (unsigned long)(255.99*frgba[i][0]) +
						(unsigned long)256 * ((unsigned long)(255.99*frgba[i][1]) +
						(unsigned long)256 * ((unsigned long)(255.99*frgba[i][2]) +
						(unsigned long)256 *  (unsigned long)(255.99*frgba[i][3])		));
	  }


    }

    /* This clear is as late as possible */
	glClearAccum(0., 0., 0., 0.);
    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT|GL_ACCUM_BUFFER_BIT);

/***************************************************************************
 * INPUT: tile information                                                 *
 ***************************************************************************/
    for(i=0; i<nsub; i++) { tile_children[i][0] = -1; }
    for(i=0; i<nsub; i++) { 
      ippp = tile_parent[i];
      nccc = 0;
      while(tile_children[ippp][nccc] >= 0) { nccc++; }
      tile_children[ippp][nccc  ] =  i;
      tile_children[ippp][nccc+1] = -1;
    }
    printf("tile_children lists generated\n");
/*
    i = 0;
    while(tile_children[node0][i] >= 0) {
      iccc = tile_children[node0][i];
      printf("child of node0 = %d    maxerr = %d\n", iccc, tile_maxerr[iccc]);
      i++;
    }
*/

/*****************************************************************************
 * GENERATE: distance**2 to each tile (rendering is done farthest to nearest *
 *           under each node                                                 *
 * distance = distance from eye to closest point JUST interior to tile   .   *
 *****************************************************************************/
    fminx = fLimits[0] - eyepos[0];
    fminy = fLimits[2] - eyepos[1];
    fminz = fLimits[4] - eyepos[2];
	delx    = (fLimits[1] - fLimits[0]) / (float)Nx;
    facXtoN = (fLimits[1] - fLimits[0]) / (float)Nx;

    for(i=0; i<nsub; i++) {
      nr = tile_nrep[i];
      nxsub = tile_nxsub[i];
      nysub = tile_nysub[i];
      nzsub = tile_nzsub[i];
      ixoff = tile_ixoff[i];
      iyoff = tile_iyoff[i];
      izoff = tile_izoff[i];
	  m_iNBoundaries = m_iNBoundariesRep[nr];

//      xx = fminx + (float)(nr*(2*tile_ixoff[i]+tile_nxsub[i]))/(float)Nx;
//      yy = fminy + (float)(nr*(2*tile_iyoff[i]+tile_nysub[i]))/(float)Nx;
//      zz = fminz + (float)(nr*(2*tile_izoff[i]+tile_nzsub[i]))/(float)Nx;

      xmin = fminx + (float)((ixoff+m_iNBoundaries)*nr) * facXtoN;
      ymin = fminy + (float)((iyoff+m_iNBoundaries)*nr) * facXtoN;
      zmin = fminz + (float)((izoff+m_iNBoundaries)*nr) * facXtoN;
  
      xmax = xmin + (float)((nxsub-2*m_iNBoundaries)*nr) * facXtoN;
      ymax = ymin + (float)((nysub-2*m_iNBoundaries)*nr) * facXtoN;
      zmax = zmin + (float)((nzsub-2*m_iNBoundaries)*nr) * facXtoN;

	  xx = 0.; if(xmin > 0.) { xx = delx + xmin; }  if(xmax < 0.) { xx = delx - xmax; }
	  yy = 0.; if(ymin > 0.) { yy = delx + ymin; }  if(ymax < 0.) { yy = delx - ymax; }
	  zz = 0.; if(zmin > 0.) { zz = delx + zmin; }  if(zmax < 0.) { zz = delx - zmax; }

      tile_dist[i] = xx*xx + yy*yy + zz*zz;
      tile_did[i] = 0;
    }

/*****************************************************************************
 * GENERATE: ordered list of tiles                                           *
 *****************************************************************************/
    delxmin = 0.;
    if(viewres > 0.) { delxmin = (float)1.01 * viewres / ((float)Nx/(fLimits[1] - fLimits[0])); }

    nlist = 1;
    list_tiles[0] = node0;
    need_check = 1;

    while(need_check == 1) {
      nlistnew = 0;
      need_check = 0;
      for(ilist = 0; ilist<nlist; ilist++) {
        inode = list_tiles[ilist];

        distmin = (float)sqrt(tile_dist[inode]);
		if(distmin < dnearclip) { distmin = (float)dnearclip; }

        delxtol  = distmin * angtol;
        delxsize =  (float)tile_nrep[inode] / ((float)Nx/(fLimits[1] - fLimits[0]));

        if(tile_maxerr[inode] <= tolerance || delxsize < delxtol
                                           || delxsize < delxmin) {
          list_new[nlistnew] = inode;
          nlistnew++;
        } else {
          need_check = 1;
          nccc = 0;
          while(tile_children[inode][nccc] >= 0) { nccc++; }
          isub = 1;
          while(isub >= 0) {
            distmax = -10000000.;
            isub = -1;
            for(iccc = 0; iccc < nccc; iccc++) {
              i = tile_children[inode][iccc];
              if(tile_did[i] == 0) {
                if(distmax < tile_dist[i]) { isub=i; distmax = tile_dist[i]; }
              }
            }
            if(isub >= 0) {
              list_new[nlistnew] = isub;
              nlistnew++;
              tile_did[isub] = 1;
            }
          }
        }
      }
      for(i=0; i<nlistnew; i++) { list_tiles[i] = list_new[i]; }
      nlist = nlistnew;
    }

/************************************************************************
 * THIS IS A TEST OF IO SPEED:  READ IN ORDER                           *
 ************************************************************************/
//	if(viewres < 1.5) {
//		nlist = 0;
//		for(i=0; i<nsub; i++) { if(tile_nrep[i] == 1) { list_tiles[nlist] = i;  nlist++; }}
//	}

/************************************************************************
 * Initialize statistics                                                *
 ************************************************************************/
	if(strcmp(m_sStats, "cross") == 0) {
		for(i=0; i<256*256; i++) { iCross[i] = 0; }
	}
	if(strcmp(m_sStats, "pdf") == 0) {
		for(i=0; i<256; i++) { iPDF[i] = 0; }
	}

/***************************************************************************
 *    hvr: subroutine "subvolrend" starts here                             *
 *    INPUTS: ixoff, iyoff, izoff, nxsub, nysub, nzsub, Nx, Ny, Nz         *
 *                                                                         *
 *    ixoff = x-offset of subvolume from left/min-edge of full volume      *
 *    iyoff = y-offset of subvolume from left/min-edge of full volume      *
 *    izoff = z-offset of subvolume from left/min-edge of full volume      *
 *    nxsub = x-dim size of sub-volume (not nec. power of 2, but <= 256)   *
 *    nysub = y-dim size of sub-volume (not nec. power of 2, but <= 256)   *
 *    nzsub = z-dim size of sub-volume (not nec. power of 2, but <= 256)   *
 *    Nx    = x-dim of full volume (not nec. power of 2, unlimitted size)  *
 *    Ny    = y-dim of full volume (not nec. power of 2, unlimitted size)  *
 *    Nz    = z-dim of full volume (not nec. power of 2, unlimitted size)  *
 *                                                                         *
 *    NOTE: ALL INPUTS ARE IN ZONES (int); ALL ZONES ARE CUBES             *
 ***************************************************************************/

/***************************************************************************
 *    CALCULATE: size and position of subvolume (in modelview space)       *
 *    CONVENTIONS: 1) x-dim always goes from (-1, 1);                      *
 *                 2) all zones are cubes                                  *
 *    ==>  y-dim goes from (-Ny/Nx, +Ny/Nx)                                *
 *    ==>  z-dim goes from (-Nz/Nx, +Nz/Nx)                                *
 *                                                                         *
 *    fulminpos[3] = xyz coordinates of min-vertex of full volume          *
 *    subminpos[3] = xyz coordinates of min-vertex of sub volume           *
 *    subvolsiz[3] = xyz size of subvolume                                 *
 *    subtexsiz[3] = xyz size of power of 2 texture subvolume in which     *
 *                   input subvolume is imbedded                           *
 *                                                                         *
 *                                                                         *
 *    +------------------------+ -  +Ny/Nx                                 *
 *    |                        | ^                                         *
 *    |                        | |                                         *
 *    |                        | |                                         *
 *    |    |<--- A --->|       | |     A = subtexsiz[0] = 2**n always      *
 *    |                        | |     B = subtexsiz[1] = 2**n always      *
 *    |    +-----------+ -     | |     C = subvolsiz[0]                    *
 *    |    |           | ^     | |     D = subvolsiz[1]                    *
 *    |    +-------+   | |     |       E = subminpos[0]                    *
 *    |    |   C   |   |       | Ny    F = subminpos[1]                    *
 *    |    |       |   | B     |       G = fulminpos[0] = -1 always        *
 *    |    |      D|   |       | |     H = fulminpos[1] = -Ny/Nx always    *
 *    |    |       |   | |     | |                                         *
 *    |    |       |   | v     | |                                         *
 *    |    +-------+---+ -     | |                                         *
 *    |  (E,F)                 | |                                         *
 *    |                        | |                                         *
 *    |                        | v                                         *
 *    +------------------------+ -  -Ny/Nx                                 *
 *  (G,H)                                                                  *
 *    |<--------- Nx --------->|                                           *
 *   -1                       +1                                           *
 *                                                                         *
 ***************************************************************************/


/**************************************************************************************
 *  Cache IO: loop over tiles in this list, request IO for those in viewing frustum   *
 **************************************************************************************/
	if(nlist > 1) { isubcur = -1; }  // Only reuse current tile of single tile renderings
	ResetCacheStatus();
	iNTileRender = 0;
    for(jsub=0; jsub<nlist; jsub++) {
	  if(iStopRendering == 1) {
			iStopRendering = 0;
			while(iIO_Read != 0) { _sleep(10); }
			return;
	  }
      isub = list_tiles[jsub];

      nr = tile_nrep[isub];
      nxsub = tile_nxsub[isub];
      nysub = tile_nysub[isub];
      nzsub = tile_nzsub[isub];
      ixoff = tile_ixoff[isub];
      iyoff = tile_iyoff[isub];
      izoff = tile_izoff[isub];
      offset64t = tile_off64t[isub];
      nbytes    = nxsub * nysub * nzsub;
//	  m_iNBoundaries = m_iNBoundariesRep[nr];
	  m_iNBoundaries = m_iNBoundariesRep[1];

      fulminpos[0] = fLimits[0];
      fulminpos[1] = fLimits[2];
      fulminpos[2] = fLimits[4];

      facXtoN = (fLimits[1]-fLimits[0]) / (float)Nx;

      subminpos[0] = fulminpos[0] + (float)((ixoff+m_iNBoundaries)*nr) * facXtoN;
      subminpos[1] = fulminpos[1] + (float)((iyoff+m_iNBoundaries)*nr) * facXtoN;
      subminpos[2] = fulminpos[2] + (float)((izoff+m_iNBoundaries)*nr) * facXtoN;
  
      subvolsiz[0] = (float)((nxsub-2*m_iNBoundaries)*nr) * facXtoN;
      subvolsiz[1] = (float)((nysub-2*m_iNBoundaries)*nr) * facXtoN;
      subvolsiz[2] = (float)((nzsub-2*m_iNBoundaries)*nr) * facXtoN;

	  m_iNBoundaries = m_iNBoundariesRep[nr];

      for(i=0; i<3; i++) { submaxpos[i] = subminpos[i] + subvolsiz[i]; }

	  // Generate Texture mem dims s.t.: smallest power of 2 just containing subtile
      nxsb2 = 2;   while(nxsb2 < nxsub) { nxsb2 *= 2; }
      nysb2 = 2;   while(nysb2 < nysub) { nysb2 *= 2; }
      nzsb2 = 2;   while(nzsb2 < nzsub) { nzsb2 *= 2; }
	  iTexBndry = 0;

	  // Use texture boundary of 1, if applicable
	  if(nxsb2/2 == nxsub-2  &&  nysb2/2 == nysub-2  &&  nzsb2/2 == nzsub-2) {
		  // In this case, can use iTexBndry = 1 with a perfect fit
		  iTexBndry = 1;
		  nxsb2 = nxsub;
		  nysb2 = nysub;
		  nzsb2 = nzsub;
	  }

     /*********************************************************************
      * START of ip loop -- LOOP over pannels                             *
      *********************************************************************/
      for(ip=0; ip < npannel; ip++) {

//	This was here, but not used for getting list of brick to read in
//  Completely changed for rendering, see rendering loop below
//			  idir = pan_idir[ip];
//			  dxx = pan_eye[ip][0];
//			  dyy = pan_eye[ip][1];
//			  dzz = pan_eye[ip][2];

		 /*********************************************************************
		  * Derive min and max distances for edges of each tile               *
		  * As a byproduct, determine if tile is OUTSIDE the viewing frustum  *
		  * A tile is OUTSIDE of a convex faceted volume if all 8 vertices    *
		  * of the tile are on the outer side of any one plane tangent to     *
		  * a facet of that volume.  Viewing frustums are convex.             *
		  *********************************************************************/
		  edgemin = 10000000.;
		  edgemax = 0.;
		  isoutside_n = 1;  /* near   */
		  isoutside_f = 1;  /* far    */
		  isoutside_t = 1;  /* top    */
		  isoutside_b = 1;  /* bottom */
		  isoutside_l = 1;  /* left   */
		  isoutside_r = 1;  /* right  */
		  for(i=0; i<2; i++) { xx = subminpos[0]+(float)i*subvolsiz[0]-eyepos[0];
		  for(j=0; j<2; j++) { yy = subminpos[1]+(float)j*subvolsiz[1]-eyepos[1];
		  for(k=0; k<2; k++) { zz = subminpos[2]+(float)k*subvolsiz[2]-eyepos[2];
			edgedist = (xx*pan_eye[ip][0]+yy*pan_eye[ip][1]+zz*pan_eye[ip][2]);
			if(edgemin > edgedist) { edgemin = edgedist; }
			if(edgemax < edgedist) { edgemax = edgedist; }

			if(edgedist > m_ShvrUse.fNearClip - softdistance   ) { isoutside_n = 0; }
			if(edgedist < m_ShvrUse.fFarrClip + softdistance   ) { isoutside_f = 0; }
			if(xx*pan_top_o[ip][0]+yy*pan_top_o[ip][1]
								  +zz*pan_top_o[ip][2] < 0. ) { isoutside_t = 0; }
			if(xx*pan_bot_o[ip][0]+yy*pan_bot_o[ip][1]
								  +zz*pan_bot_o[ip][2] < 0. ) { isoutside_b = 0; }
			if(xx*pan_lef_o[ip][0]+yy*pan_lef_o[ip][1]
								  +zz*pan_lef_o[ip][2] < 0. ) { isoutside_l = 0; }
			if(xx*pan_rit_o[ip][0]+yy*pan_rit_o[ip][1]
								  +zz*pan_rit_o[ip][2] < 0. ) { isoutside_r = 0; }
		  }}}

		/******************************************************************************
		 *    Test for Object oriented clipping planes                                *
		 ******************************************************************************/
		  if(iTexBndry == 1) {
				xx0 = (float)0.0;
				xx1 = (float)1.0;
				yy0 = (float)0.0;
				yy1 = (float)1.0;
				zz0 = (float)0.0;
				zz1 = (float)1.0;
		  } else {
				xx0 =                 (float)1.0  / (float)(nxsb2-1);
				xx1 = ((float)nxsub - (float)2.0) / (float)(nxsb2-1);
				yy0 =                 (float)1.0  / (float)(nysb2-1);
				yy1 = ((float)nysub - (float)2.0) / (float)(nysb2-1);
				zz0 =                 (float)1.0  / (float)(nzsb2-1);
				zz1 = ((float)nzsub - (float)2.0) / (float)(nzsb2-1);
		  }

		  iClipThis0 = 0;
		  px0 = subminpos[0];
		  px1 = submaxpos[0];
		  py0 = subminpos[1];
		  py1 = submaxpos[1];
		  pz0 = subminpos[2];
		  pz1 = submaxpos[2];
		  if(m_ShvrUse.ifauxclip[0] == 2) {
				px0c = m_ShvrUse.auxclip[0][3];
				px1c = m_ShvrUse.auxclip[1][3];
				py0c = m_ShvrUse.auxclip[2][3];
				py1c = m_ShvrUse.auxclip[3][3];
				pz0c = m_ShvrUse.auxclip[4][3];
				pz1c = m_ShvrUse.auxclip[5][3];
				xx0c = xx0 + (xx1-xx0)*(px0c-subminpos[0])/subvolsiz[0];
				xx1c = xx0 + (xx1-xx0)*(px1c-subminpos[0])/subvolsiz[0];
				yy0c = yy0 + (yy1-yy0)*(py0c-subminpos[1])/subvolsiz[1];
				yy1c = yy0 + (yy1-yy0)*(py1c-subminpos[1])/subvolsiz[1];
				zz0c = zz0 + (zz1-zz0)*(pz0c-subminpos[2])/subvolsiz[2];
				zz1c = zz0 + (zz1-zz0)*(pz1c-subminpos[2])/subvolsiz[2];
				if(xx0c > xx0) { xx0 = xx0c;    px0 = px0c; }
				if(xx1c < xx1) { xx1 = xx1c;    px1 = px1c; }
				if(yy0c > yy0) { yy0 = yy0c;    py0 = py0c; }
				if(yy1c < yy1) { yy1 = yy1c;    py1 = py1c; }
				if(zz0c > zz0) { zz0 = zz0c;    pz0 = pz0c; }
				if(zz1c < zz1) { zz1 = zz1c;    pz1 = pz1c; }
				if(xx0 >= xx1) { iClipThis0 = 1; }
				if(yy0 >= yy1) { iClipThis0 = 1; }
				if(zz0 >= zz1) { iClipThis0 = 1; }
		  }

		/********************************************************************************
		 *    SKIP this tile if it is outside viewing frustum OR object oriented clips  *
		 *    this is end of jsub (tile) loop if tile is outside                        *
		 ********************************************************************************/
		  if(   isoutside_n == 1 || isoutside_f == 1 || isoutside_t == 1
			 || isoutside_b == 1 || isoutside_l == 1 || isoutside_r == 1 || iClipThis0 == 1) {

		  } else {
			// Submit request to read in this tile
			iTileRender[iNTileRender] = isub;
			iNTileRender++;
			if(isubcur != isub) {
				sprintf(tile_full, "%s%s\0", tile_path, tile_file[isub]);
				RequestCacheIO(isub, tile_full, offset64t, (DWORD)nbytes);
			}
		  }
	  }  // End of ipannel loop
	}    // End of jsub loop

	// All read requests are in, set IO thread going
	iIO_Read = 1;

/*********************************************************************************
 *  RENDER: loop over tiles in this list and render those in te viewing frustum  *
 *********************************************************************************/
    ntiles  = 0;
    nslices = 0;
//    for(jsub=0; jsub<nlist; jsub++)
    for(jsub=0; jsub<iNTileRender; jsub++) {

	  if(m_iTimersOn == 1) { tsEvent.StampTime("Begin jsubRenderloop"); }

		if(iStopRendering == 1) {
			iStopRendering = 0;
			isubcur = -1;
			while(iIO_Read != 0) { _sleep(10); }
			return;
		}
//      isub = list_tiles[jsub];
	  isub = iTileRender[jsub];

      nr = tile_nrep[isub];
      nxsub = tile_nxsub[isub];
      nysub = tile_nysub[isub];
      nzsub = tile_nzsub[isub];
      ixoff = tile_ixoff[isub];
      iyoff = tile_iyoff[isub];
      izoff = tile_izoff[isub];
      offset64t = tile_off64t[isub];
      nbytes    = nxsub * nysub * nzsub;
//	  m_iNBoundaries = m_iNBoundariesRep[nr];
	  m_iNBoundaries = m_iNBoundariesRep[1];

      fulminpos[0] = fLimits[0];
      fulminpos[1] = fLimits[2];
      fulminpos[2] = fLimits[4];

      facXtoN = (fLimits[1]-fLimits[0]) / (float)Nx;

      subminpos[0] = fulminpos[0] + (float)((ixoff+m_iNBoundaries)*nr) * facXtoN;
      subminpos[1] = fulminpos[1] + (float)((iyoff+m_iNBoundaries)*nr) * facXtoN;
      subminpos[2] = fulminpos[2] + (float)((izoff+m_iNBoundaries)*nr) * facXtoN;
  
      subvolsiz[0] = (float)((nxsub-2*m_iNBoundaries)*nr) * facXtoN;
      subvolsiz[1] = (float)((nysub-2*m_iNBoundaries)*nr) * facXtoN;
      subvolsiz[2] = (float)((nzsub-2*m_iNBoundaries)*nr) * facXtoN;

	  m_iNBoundaries = m_iNBoundariesRep[nr];

      for(i=0; i<3; i++) { submaxpos[i] = subminpos[i] + subvolsiz[i]; }

	  // Generate Texture mem dims s.t.: smallest power of 2 just containing subtile
      nxsb2 = 2;   while(nxsb2 < nxsub) { nxsb2 *= 2; }
      nysb2 = 2;   while(nysb2 < nysub) { nysb2 *= 2; }
      nzsb2 = 2;   while(nzsb2 < nzsub) { nzsb2 *= 2; }
	  iTexBndry = 0;

	  // Use texture boundary of 1, if applicable
	  if(nxsb2/2 == nxsub-2  &&  nysb2/2 == nysub-2  &&  nzsb2/2 == nzsub-2) {
		  // In this case, can use iTexBndry = 1 with a perfect fit
		  iTexBndry = 1;
		  nxsb2 = nxsub;
		  nysb2 = nysub;
		  nzsb2 = nzsub;
	  }

     /*********************************************************************
      * START of ip loop -- LOOP over pannels                             *
      *********************************************************************/
      for(ip=0; ip < npannel; ip++) {

		  // Generate rendering plane orientation
		  // Function of view frustum and sub-brick
		  // Construct vector from eye to centroid of that part of the brick which is
		  //     inside the view frustum.
		  // To do this 1) project verices of brick onto window at unit distance 
		  //            2) crop projection to view frustum
		  //            3) generate centroid of cropped image
		  //            4) base idir on max component of this eye to centroid vector

		  // For each vertice of brick: get horiz & vert limits of bounding box
		  // on a window which is at unit distance along panel normal
/*
		  vhhmin = (float)1.0;
		  vhhmax = (float)0.0;
		  isBehind = 1;
		  for(i=0; i<2; i++) { xx = subminpos[0]+(float)i*subvolsiz[0]-eyepos[0];
		  for(j=0; j<2; j++) { yy = subminpos[1]+(float)j*subvolsiz[1]-eyepos[1];
		  for(k=0; k<2; k++) { zz = subminpos[2]+(float)k*subvolsiz[2]-eyepos[2];
			vhh = xx*pan_torq[ip][0] +yy*pan_torq[ip][1] +zz*pan_torq[ip][2]; 
			vnn = xx* pan_eye[ip][0] +yy* pan_eye[ip][1] +zz* pan_eye[ip][2]; 
			vvv = xx*  pan_up[ip][0] +yy*  pan_up[ip][1] +zz*  pan_up[ip][2]; 

			if(vnn > m_ShvrUse.fNearClip - softdistance) {
				// Found vertex which is in front of you
				isBehind = 0;

				// Input vertex into limits of bounding box
				vhh = vhh / vnn;
				vvv = vvv / vnn;
				if(vhhmin > vhhmax) {
					vhhmin = vhh;
					vhhmax = vhh;
					vvvmin = vvv;
					vvvmax = vvv;
				} else {
					if(vhhmin > vhh) { vhhmin = vhh; }
					if(vhhmax < vhh) { vhhmax = vhh; }
					if(vvvmin > vvv) { vvvmin = vvv; }
					if(vvvmax < vvv) { vvvmax = vvv; }
				}
			}
		  }}}

		  if(isBehind == 1) {
			  // If no vertex was in front of you, just use panel info
			  idir = pan_idir[ip];
			  dxx = pan_eye[ip][0];
			  dyy = pan_eye[ip][1];
			  dzz = pan_eye[ip][2];
		  } else {
			  // Crop limits of bounding box to view frustum
			  if(vhhmin < pan_dlef[ip]) { vhhmin = pan_dlef[ip]; }
			  if(vhhmax > pan_drit[ip]) { vhhmax = pan_drit[ip]; }
			  if(vvvmin < pan_dbot[ip]) { vvvmin = pan_dbot[ip]; }
			  if(vvvmax > pan_dtop[ip]) { vvvmax = pan_dtop[ip]; }

			  // Generate normalized eye to centroid vector
			  vhhcent = (float)0.5 * (vhhmin + vhhmax);
			  vvvcent = (float)0.5 * (vvvmin + vvvmax);
			  dxx = pan_eye[ip][0] + vhhcent*pan_torq[ip][0] + vvvcent*pan_up[ip][0];
			  dyy = pan_eye[ip][1] + vhhcent*pan_torq[ip][1] + vvvcent*pan_up[ip][1];
			  dzz = pan_eye[ip][2] + vhhcent*pan_torq[ip][2] + vvvcent*pan_up[ip][2];

			  vnormi = (float)(1.0 / sqrt(dxx*dxx + dyy*dyy + dzz*dzz));
			  dxx = dxx * vnormi;
			  dyy = dyy * vnormi;
			  dzz = dzz * vnormi;
		  }

*/
		  // Direction to center of subvolume
		  dxx = subminpos[0]+(float)0.5*subvolsiz[0]-eyepos[0];
		  dyy = subminpos[1]+(float)0.5*subvolsiz[1]-eyepos[1];
		  dzz = subminpos[2]+(float)0.5*subvolsiz[2]-eyepos[2];

		  // Direction to center of subvolume in coordinate attached to panel
		  vhh = dxx*pan_torq[ip][0] + dyy*pan_torq[ip][1] + dzz*pan_torq[ip][2]; 
		  vnn = dxx* pan_eye[ip][0] + dyy* pan_eye[ip][1] + dzz* pan_eye[ip][2]; 
		  vvv = dxx*  pan_up[ip][0] + dyy*  pan_up[ip][1] + dzz*  pan_up[ip][2]; 

		  if(vnn <= (float)0.0) {
			  // Use panel normal if center of cube is behind your head
			  dxx = pan_eye[ip][0];
			  dyy = pan_eye[ip][1];
			  dzz = pan_eye[ip][2];
		  } else {
			  // Limit direction to be within view frustum
			  vhh = vhh / vnn;
			  vvv = vvv / vnn;
			  if(vhh < pan_dlef[ip]) { vhh = pan_dlef[ip]; }
			  if(vhh > pan_drit[ip]) { vhh = pan_drit[ip]; }
			  if(vvv < pan_dbot[ip]) { vvv = pan_dbot[ip]; }
			  if(vvv > pan_dtop[ip]) { vvv = pan_dtop[ip]; }
			  dxx = pan_eye[ip][0] + vhh*pan_torq[ip][0] + vvv*pan_up[ip][0];
			  dyy = pan_eye[ip][1] + vhh*pan_torq[ip][1] + vvv*pan_up[ip][1];
			  dzz = pan_eye[ip][2] + vhh*pan_torq[ip][2] + vvv*pan_up[ip][2];

			  // Normalize resulting direction
			  vnormi = (float)(1.0 / sqrt(dxx*dxx + dyy*dyy + dzz*dzz));
			  dxx = dxx * vnormi;
			  dyy = dyy * vnormi;
			  dzz = dzz * vnormi;
		  }

		  // Base idir on max absolute value of coordinate of bounded center-of-subvolume vector
		  if(dxx > 0.) { axx = dxx; } else { axx = -dxx; }
		  if(dyy > 0.) { ayy = dyy; } else { ayy = -dyy; }
		  if(dzz > 0.) { azz = dzz; } else { azz = -dzz; }
		  idir = 2;                                     // view along z-direction
		  if(ayy > axx  &&  ayy > azz) { idir = 1; }    // view along y-direction
		  if(axx > ayy  &&  axx > azz) { idir = 0; }    // view along x-direction

     /*********************************************************************
      * Derive min and max distances for edges of each tile               *
      * As a byproduct, determine if tile is OUTSIDE the viewing frustum  *
      * A tile is OUTSIDE of a convex faceted volume if all 8 vertices    *
      * of the tile are on the outer side of any one plane tangent to     *
      * a facet of that volume.  Viewing frustums are convex.             *
      *********************************************************************/
      edgemin = 10000000.;
      edgemax = 0.;
      isoutside_n = 1;  /* near   */
      isoutside_f = 1;  /* far    */
      isoutside_t = 1;  /* top    */
      isoutside_b = 1;  /* bottom */
      isoutside_l = 1;  /* left   */
      isoutside_r = 1;  /* right  */
      for(i=0; i<2; i++) { xx = subminpos[0]+(float)i*subvolsiz[0]-eyepos[0];
      for(j=0; j<2; j++) { yy = subminpos[1]+(float)j*subvolsiz[1]-eyepos[1];
      for(k=0; k<2; k++) { zz = subminpos[2]+(float)k*subvolsiz[2]-eyepos[2];
        edgedist = (xx*pan_eye[ip][0]+yy*pan_eye[ip][1]+zz*pan_eye[ip][2]);
        if(edgemin > edgedist) { edgemin = edgedist; }
        if(edgemax < edgedist) { edgemax = edgedist; }

        if(edgedist > m_ShvrUse.fNearClip - softdistance   ) { isoutside_n = 0; }
        if(edgedist < m_ShvrUse.fFarrClip + softdistance   ) { isoutside_f = 0; }
        if(xx*pan_top_o[ip][0]+yy*pan_top_o[ip][1]
                              +zz*pan_top_o[ip][2] < 0. ) { isoutside_t = 0; }
		if(xx*pan_bot_o[ip][0]+yy*pan_bot_o[ip][1]
                              +zz*pan_bot_o[ip][2] < 0. ) { isoutside_b = 0; }
        if(xx*pan_lef_o[ip][0]+yy*pan_lef_o[ip][1]
                              +zz*pan_lef_o[ip][2] < 0. ) { isoutside_l = 0; }
        if(xx*pan_rit_o[ip][0]+yy*pan_rit_o[ip][1]
                              +zz*pan_rit_o[ip][2] < 0. ) { isoutside_r = 0; }
      }}}

/******************************************************************************
 *    Test for Object oriented clipping planes                                *
 ******************************************************************************/
	  if(iTexBndry == 1) {
			xx0 = (float)0.0;
			xx1 = (float)1.0;
			yy0 = (float)0.0;
			yy1 = (float)1.0;
			zz0 = (float)0.0;
			zz1 = (float)1.0;
	  } else {
			xx0 =                 (float)1.0  / (float)(nxsb2-1);
			xx1 = ((float)nxsub - (float)2.0) / (float)(nxsb2-1);
			yy0 =                 (float)1.0  / (float)(nysb2-1);
			yy1 = ((float)nysub - (float)2.0) / (float)(nysb2-1);
			zz0 =                 (float)1.0  / (float)(nzsb2-1);
			zz1 = ((float)nzsub - (float)2.0) / (float)(nzsb2-1);
	  }

	  iClipThis0 = 0;
	  px0 = subminpos[0];
	  px1 = submaxpos[0];
	  py0 = subminpos[1];
	  py1 = submaxpos[1];
	  pz0 = subminpos[2];
	  pz1 = submaxpos[2];
	  if(m_ShvrUse.ifauxclip[0] == 2) {
			px0c = m_ShvrUse.auxclip[0][3];
			px1c = m_ShvrUse.auxclip[1][3];
			py0c = m_ShvrUse.auxclip[2][3];
			py1c = m_ShvrUse.auxclip[3][3];
			pz0c = m_ShvrUse.auxclip[4][3];
			pz1c = m_ShvrUse.auxclip[5][3];
			xx0c = xx0 + (xx1-xx0)*(px0c-subminpos[0])/subvolsiz[0];
			xx1c = xx0 + (xx1-xx0)*(px1c-subminpos[0])/subvolsiz[0];
			yy0c = yy0 + (yy1-yy0)*(py0c-subminpos[1])/subvolsiz[1];
			yy1c = yy0 + (yy1-yy0)*(py1c-subminpos[1])/subvolsiz[1];
			zz0c = zz0 + (zz1-zz0)*(pz0c-subminpos[2])/subvolsiz[2];
			zz1c = zz0 + (zz1-zz0)*(pz1c-subminpos[2])/subvolsiz[2];
			if(xx0c > xx0) { xx0 = xx0c;    px0 = px0c; }
			if(xx1c < xx1) { xx1 = xx1c;    px1 = px1c; }
			if(yy0c > yy0) { yy0 = yy0c;    py0 = py0c; }
			if(yy1c < yy1) { yy1 = yy1c;    py1 = py1c; }
			if(zz0c > zz0) { zz0 = zz0c;    pz0 = pz0c; }
			if(zz1c < zz1) { zz1 = zz1c;    pz1 = pz1c; }
			if(xx0 >= xx1) { iClipThis0 = 1; }
			if(yy0 >= yy1) { iClipThis0 = 1; }
			if(zz0 >= zz1) { iClipThis0 = 1; }
	  }

/********************************************************************************
 *    SKIP this tile if it is outside viewing frustum OR object oriented clips  *
 *    this is end of jsub (tile) loop if tile is outside                        *
 ********************************************************************************/
      if(   isoutside_n == 1 || isoutside_f == 1 || isoutside_t == 1
         || isoutside_b == 1 || isoutside_l == 1 || isoutside_r == 1 || iClipThis0 == 1) {

      } else {

     /************************************************************
      * Read in subtile isub using DIRECT IO (if not already in) *
      ************************************************************/
	  if(isubcur != isub) {
		ntiles++;

		// Wait for IO to finish and get pointer to buffer
		if(m_iTimersOn == 1) { tsEvent.StampTime("Begin GetCacheBuffer"); }
		bobsubZ = (unsigned char *)lpGetCacheBuffer(isub);
		if(m_iTimersOn == 1) { sprintf(str, "End   GetCacheBuffer : %d bytes", nzsub*nysub*nxsub);
							   tsEvent.StampTime(str); }

		// Most fast implementations of OpenGL require power of 2 dimensions
		// If tile is NOT a power of 2 in x or y:
		//   1) suffle memory pointers
		//   2) copy into a format which IS power of 2 in x and y
		if(nxsb2 != nxsub  ||  nysb2 != nysub) {
			if(m_iTimersOn == 1) { tsEvent.StampTime("Begin OddBrickFill"); }
			bobin = bobsubZ;
			bobsubZ = bobsubZ0;
			ic = 0;
			for(k=0; k<nzsub; k++) {
			for(j=0; j<nysub; j++) { joff = nxsb2*(j + nysb2 * k);
			for(i=0; i<nxsub; i++) {
				bobsubZ[i + joff] = bobin[ic];
				ic++;
			}}}
			if(m_iTimersOn == 1) { sprintf(str, "End   OddBrickFill : %d bytes", nzsub*nysub*nxsub);
									   tsEvent.StampTime(str); }
		}

        sprintf(tile_full, "%s%s\0", tile_path, tile_file[isub]);
		if(iOpac == 1) {
			sprintf(tile_fileO, "%s", tile_file[isub]);
			tile_fileO[ 3] = 'A';
			tile_fileO[ 4] = 'L';
			tile_fileO[ 5] = 'P';
			tile_fileO[ 6] = 'H';
			tile_fileO[ 7] = 'A';
			tile_fileO[ 8] = '_';
			tile_fileO[ 9] = '_';
			tile_fileO[10] = '_';
			sprintf(m_sTileO, "%s%s\0", tile_path, tile_fileO);
		}

		if(iOpac == 1  ||  strcmp(m_sStats, "cross") == 0) {
			if(nxsb2 == nxsub  &&  nysb2 == nysub) {
				if(m_iBufferedIO == 1) { dwGot = dhpReadFromFileNoBuf(bobOZ, m_sTileO, offset64t, (DWORD)nbytes); }
				else                   { dwGot = dhpReadFromFileSync( bobOZ, m_sTileO, offset64t, (DWORD)nbytes); }
			} else {
				if(m_iBufferedIO == 1) { dwGot = dhpReadFromFileNoBuf(bobOY, m_sTileO, offset64t, (DWORD)nbytes); } 
				else                   { dwGot = dhpReadFromFileSync( bobOY, m_sTileO, offset64t, (DWORD)nbytes); }
				ic = 0;
				for(k=0; k<nzsub; k++) {
				for(j=0; j<nysub; j++) {
				for(i=0; i<nxsub; i++) {
					bobOZ[i + nxsb2*(j + nysb2 * k)] = bobOY[ic];
					ic++;
				}}}
			}
		}

		bSetBobX = FALSE;
		bSetBobY = FALSE;
        isubcur = isub;
        ireload = 1;
      }
/*****************************************************************************
 * BEGIN multiple pannel stuff                                               *
 *****************************************************************************/
      if(ip != ipcur) {

       /****************************************************
        *  set up frustum for pannel ip                    *
        ****************************************************/
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();

        dleft   = dnearclip * pan_dlef[ip];
        dright  = dnearclip * pan_drit[ip];
        dbottom = dnearclip * pan_dbot[ip];
        dtop    = dnearclip * pan_dtop[ip];
        glFrustum( dleft, dright, dbottom, dtop, dnearclip, dfarclip );
        gluLookAt(m_ShvrUse.fEye[0], m_ShvrUse.fEye[1], m_ShvrUse.fEye[2],
                  pan_lookat[ip][0], pan_lookat[ip][1], pan_lookat[ip][2],
                  pan_up[ip][0],     pan_up[ip][1],     pan_up[ip][2]);

       /*************************************************************
        *  swap out old pannel (if nec.) and swap in new pannel ip  *
        *************************************************************/
        glMatrixMode(GL_TEXTURE);
        glDisable(DHP_TEX);
        glDisable(GL_BLEND);
        if(ipcur > -1) {
			glReadPixels(0,0,m_ShvrUse.iWidth,m_ShvrUse.iHeight,
				GL_RGB,GL_UNSIGNED_BYTE,rgbpannel[ipcur]);
        }

        glDrawPixels(m_ShvrUse.iWidth, m_ShvrUse.iHeight,
			GL_RGB,GL_UNSIGNED_BYTE,rgbpannel[ip]);
        ipcur = ip;
        glEnable(GL_BLEND);
        glEnable(DHP_TEX);
	  }
/*****************************************************************************
 * END   multiple pannel stuff                                               *
 *****************************************************************************/

     /***********************************************************************
      *  SET scale and offset of texture memory for subvolume               *
      *  ASSUMES glLoadIdentity sets limits of                              *
      *          --> subtexture <--  to (0,1)x(0,1)x(0,1)                   *
      *          which get mapped to                                        *
      *          (subminpos[i], subminpos[i]+subvolsiz[i])x for(i=0,1,2)    *
      ***********************************************************************/

      while ( glGetError() != GL_NO_ERROR ) ;

     /*********************************************************************
      *    THIS IS WHERE ALL ALPHA BLENDING IS DONE                       *
      *********************************************************************/

	 /************************************************************************
      * Collect statistics                                                   *
      ************************************************************************/
	  if(strcmp(m_sStats, "cross") == 0) {
			for(k=0; k<nzsub; k++) {
			for(j=0; j<nysub; j++) {
			for(i=0; i<nxsub; i++) {
				ijk = i + nxsb2*(j + nysb2 * k);
				iCross[(int)bobsubZ[ijk] + 256*(int)bobOZ[ijk]]++;
			}}}
	  }
	  if(strcmp(m_sStats, "pdf") == 0) {
			for(k=0; k<nzsub; k++) {
			for(j=0; j<nysub; j++) {
			for(i=0; i<nxsub; i++) {
				iPDF[(int)bobsubZ[i + nxsb2*(j + nysb2 * k)]]++;
			}}}
	  }

	 /************************************************************************
      * Calculate vector from eye to brick for correct rendering orientation *
      ************************************************************************/
	  glBindTexture(DHP_TEX, texname[1]);

	 /**************************************************************
	  * Rendering direction is along X-axis                        *
	  **************************************************************/
	  if(idir == 0) {
		if(bSetBobX == FALSE) {
			if(m_iTimersOn == 1) { tsEvent.StampTime("Begin TposeYZ"); }
			TposeYZ( bobsubZ, bobsubX, &nxsb2, &nysb2, &nzsb2, &nxsub, &nysub, &nzsub );
			if(iOpac == 1) {
				TposeYZ( bobOZ, bobOX, &nxsb2, &nysb2, &nzsb2, &nxsub, &nysub, &nzsub );
			}
			bSetBobX = TRUE;
			if(m_iTimersOn == 1) { tsEvent.StampTime("End   TposeYZ"); }
		}
		if(m_iTimersOn == 1) {
			if(m_bColorTableIsAvailable) { tsEvent.StampTime("Begin Bob_RenderingYZ"); }
			else                         { tsEvent.StampTime("Begin Bol_RenderingYZ"); }
		}
/*
		if(iTexBndry == 1) {
			yy0 = (float)0.0;
			yy1 = (float)1.0;
			zz0 = (float)0.0;
			zz1 = (float)1.0;
		} else {
			yy0 =                 (float)1.0  / (float)(nysb2-1);
			yy1 = ((float)nysub - (float)2.0) / (float)(nysb2-1);
			zz0 =                 (float)1.0  / (float)(nzsb2-1);
			zz1 = ((float)nzsub - (float)2.0) / (float)(nzsb2-1);
		}
*/
		if(dxx < 0.) { ids0 = 0;                                             idsd =  1; }
		else	 	 { ids0 = ((nxsub-2*m_iNBoundaries) * (int)nplanes) - 1; idsd = -1; }

	    islicesize = nysb2*nzsb2;
		dx  = subvolsiz[0]/(float)((nxsub-2*m_iNBoundaries) * nplanes);
		dx1 = subvolsiz[0]/(float) (nxsub-2*m_iNBoundaries);
// axx,ayy,azz is already normalized
//      alpha = (float)alpha0 * ((float)sqrt(axx*axx+ayy*ayy+azz*azz)/(axx)) * (dx/((float)2.*cubesz[0]));
//		alpha = (float)alpha0 * (dx/((float)2.*cubesz[0]*axx));
		alpha = (float)alpha0 * (dx/((float)2.*cubesz[0]));
        glColor4d(1.0, 1.0, 1.0, (double)alpha);

		for(islice = 0;  islice < ((nxsub-2*m_iNBoundaries) * nplanes);  islice++) {
		xxv = subminpos[0] + dx * ((float)0.5 + (float)(ids0 + idsd*islice));

		iClipThis = iClipThis0;
		if(m_ShvrUse.ifauxclip[0] == 2  &&  xxv < m_ShvrUse.auxclip[0][3]) { iClipThis = 1; }
		if(m_ShvrUse.ifauxclip[1] == 2  &&  xxv > m_ShvrUse.auxclip[1][3]) { iClipThis = 1; }
		if(iClipThis == 0) {

			islice0  = (int)((float)0.5 + (xxv - subminpos[0]) / dx1) + m_iNBoundaries - 1;
			islice1 = islice0 + 1;
			if(islice0 <      0) { islice0 = 0; }
			if(islice1 >= nxsub) { islice1 = nxsub - 1; }

			if(m_bColorTableIsAvailable  &&  iOpac == 0) {
				if(islice0 == islice1  ||  nplanes == 1) {
					isliceoff = islicesize * islice0;
				} else {
					isliceoff0 = islicesize * islice0;
					isliceoff1 = islicesize * islice1;
					ppv0 = subminpos[0] + dx1 * ((float)0.5 + (float)(islice0-m_iNBoundaries));
					ppv1 = subminpos[0] + dx1 * ((float)0.5 + (float)(islice1-m_iNBoundaries));
					wt0 = (ppv1 - xxv) / (ppv1 - ppv0);
					if(wt0 > 0.5) { isliceoff = isliceoff0; } else { isliceoff = isliceoff1; }
				}
				glTexImage2D(DHP_TEX, 0, GL_COLOR_INDEX8_EXT,
					nysb2,nzsb2,iTexBndry, GL_COLOR_INDEX, GL_UNSIGNED_BYTE,
					&bobsubX[isliceoff]);
			} else {
				if(islice0 == islice1  ||  nplanes == 1) {
					isliceoff = islicesize * islice0;
					if(iOpac == 0) {
						for(i=0; i<nysb2*nzsb2; i++) { sol[i] = m_lrgba[bobsubX[isliceoff+i]]; }
					} else {
						wt0 = (float)255.99;
						for(i=0; i<nysb2*nzsb2; i++) {
							iRGB = bobsubX[isliceoff+i];
							iALP = bobOX[isliceoff+i];
							sol[i] =  (unsigned long)(wt0*frgba[iRGB][0])
								  + (((unsigned long)(wt0*frgba[iRGB][1]))<<8 )
								  + (((unsigned long)(wt0*frgba[iRGB][2]))<<16)
								  + (((unsigned long)(wt0*frgba[iALP][3]))<<24);
						}
					}
				} else {
					isliceoff0 = islicesize * islice0;
					isliceoff1 = islicesize * islice1;
					ppv0 = subminpos[0] + dx1 * ((float)0.5 + (float)(islice0-m_iNBoundaries));
					ppv1 = subminpos[0] + dx1 * ((float)0.5 + (float)(islice1-m_iNBoundaries));
					wt0 = (float)255.99 * (ppv1 - xxv) / (ppv1 - ppv0);
					for(i=0; i<256; i++) {
						lrgba0[i] =   (unsigned long)(wt0*frgba[i][0])
								  + (((unsigned long)(wt0*frgba[i][1]))<<8 )
								  + (((unsigned long)(wt0*frgba[i][2]))<<16)
								  + (((unsigned long)(wt0*frgba[i][3]))<<24);
						lrgba1[i] = m_lrgba[i] - lrgba0[i];
					}
					for(i=0; i<nysb2*nzsb2; i++) {
						sol[i] = lrgba0[bobsubX[isliceoff0+i]] + lrgba1[bobsubX[isliceoff1+i]];
					}
				}
				glTexImage2D(DHP_TEX,0,GL_RGBA,nysb2,nzsb2,iTexBndry,
					GL_RGBA,GL_UNSIGNED_BYTE,sol);
			}

			glBegin(GL_QUADS);
				glTexCoord2f(yy0, zz0);    glVertex3f(xxv, py0, pz0);
				glTexCoord2f(yy0, zz1);    glVertex3f(xxv, py0, pz1);
				glTexCoord2f(yy1, zz1);    glVertex3f(xxv, py1, pz1);
				glTexCoord2f(yy1, zz0);    glVertex3f(xxv, py1, pz0);
			glEnd();
			nslices++;
		}}     /* end of islice:for-loop and (iClipThis == 1):if-block */

		if(m_iTimersOn == 1) {
			if(m_bColorTableIsAvailable) { tsEvent.StampTime("End   Bob_RenderingYZ"); }
			else                         { tsEvent.StampTime("End   Bol_RenderingYZ"); }
		}
	  }

	 /**************************************************************
	  * Rendering direction is along Y-axis                        *
	  **************************************************************/
	  if(idir == 1) {
		if(bSetBobY == FALSE) {
			if(m_iTimersOn == 1) { tsEvent.StampTime("Begin TposeXZ"); }
			TposeXZ( bobsubZ, bobsubY, nxsb2, nysb2, nzsb2, nxsub, nysub, nzsub );
			if(iOpac == 1) {
				TposeXZ( bobOZ, bobOY, nxsb2, nysb2, nzsb2, nxsub, nysub, nzsub  );
			}
			bSetBobY = TRUE;
			if(m_iTimersOn == 1) { tsEvent.StampTime("End   TposeXZ"); }
		}

		if(m_iTimersOn == 1) {
			if(m_bColorTableIsAvailable) { tsEvent.StampTime("Begin Bob_RenderingXZ"); }
			else                         { tsEvent.StampTime("Begin Bol_RenderingXZ"); }
		}
		if(dyy < 0.) { ids0 = 0;                                             idsd =  1; }
		else	 	 { ids0 = ((nysub-2*m_iNBoundaries) * (int)nplanes) - 1; idsd = -1; }

	    islicesize = nzsb2*nxsb2;
		dy  = subvolsiz[1]/(float)((nysub-2*m_iNBoundaries) * nplanes);
		dy1 = subvolsiz[1]/(float) (nysub-2*m_iNBoundaries);
// axx,ayy,azz is already normalized
//      alpha = (float)alpha0 * ((float)sqrt(axx*axx+ayy*ayy+azz*azz)/(ayy)) * (dy/((float)2.*cubesz[0]));
//		alpha = (float)alpha0 * (dy/((float)2.*cubesz[0]*ayy));
		alpha = (float)alpha0 * (dy/((float)2.*cubesz[0]));
        glColor4d(1.0, 1.0, 1.0, (double)alpha);

		for(islice = 0;  islice < ((nysub-2*m_iNBoundaries) * nplanes);  islice++) {
		yyv = subminpos[1] + dy * ((float)0.5 + (float)(ids0 + idsd*islice));

		iClipThis = iClipThis0;
		if(m_ShvrUse.ifauxclip[2] == 2  &&  yyv < m_ShvrUse.auxclip[2][3]) { iClipThis = 1; }
		if(m_ShvrUse.ifauxclip[3] == 2  &&  yyv > m_ShvrUse.auxclip[3][3]) { iClipThis = 1; }
		if(iClipThis == 0) {

			islice0  = (int)((float)0.5 + (yyv - subminpos[1]) / dy1) + m_iNBoundaries - 1;
			islice1 = islice0 + 1;
			if(islice0 <      0) { islice0 = 0; }
			if(islice1 >= nysub) { islice1 = nysub - 1; }

			if(m_bColorTableIsAvailable  &&  iOpac == 0) {
				if(islice0 == islice1  ||  nplanes == 1) {
					isliceoff = islicesize * islice0;
				} else {
					isliceoff0 = islicesize * islice0;
					isliceoff1 = islicesize * islice1;
					ppv0 = subminpos[1] + dy1 * ((float)0.5 + (float)(islice0-m_iNBoundaries));
					ppv1 = subminpos[1] + dy1 * ((float)0.5 + (float)(islice1-m_iNBoundaries));
					wt0 = (ppv1 - yyv) / (ppv1 - ppv0);
					if(wt0 > 0.5) { isliceoff = isliceoff0; } else { isliceoff = isliceoff1; }
				}
				glTexImage2D(DHP_TEX, 0, GL_COLOR_INDEX8_EXT,
					nxsb2,nzsb2,iTexBndry, GL_COLOR_INDEX, GL_UNSIGNED_BYTE,
					&bobsubY[isliceoff]);
			} else {
				if(islice0 == islice1  ||  nplanes == 1) {
					isliceoff = islicesize * islice0;
					if(iOpac == 0) {
						for(i=0; i<nxsb2*nzsb2; i++) { sol[i] = m_lrgba[bobsubY[isliceoff+i]]; }
					} else {
						wt0 = (float)255.99;
						for(i=0; i<nxsb2*nxsb2; i++) {
							iRGB = bobsubY[isliceoff+i];
							iALP = bobOY[isliceoff+i];
							sol[i] =  (unsigned long)(wt0*frgba[iRGB][0])
								  + (((unsigned long)(wt0*frgba[iRGB][1]))<<8 )
								  + (((unsigned long)(wt0*frgba[iRGB][2]))<<16)
								  + (((unsigned long)(wt0*frgba[iALP][3]))<<24);
						}
					}
				} else {
					isliceoff0 = islicesize * islice0;
					isliceoff1 = islicesize * islice1;
					ppv0 = subminpos[1] + dy1 * ((float)0.5 + (float)(islice0-m_iNBoundaries));
					ppv1 = subminpos[1] + dy1 * ((float)0.5 + (float)(islice1-m_iNBoundaries));
					wt0 = (float)255.99 * (ppv1 - yyv) / (ppv1 - ppv0);
					for(i=0; i<256; i++) {
						lrgba0[i] =   (unsigned long)(wt0*frgba[i][0])
								  + (((unsigned long)(wt0*frgba[i][1]))<<8 )
								  + (((unsigned long)(wt0*frgba[i][2]))<<16)
								  + (((unsigned long)(wt0*frgba[i][3]))<<24);
						lrgba1[i] = m_lrgba[i] - lrgba0[i];
					}

					for(i=0; i<nxsb2*nzsb2; i++) {
						sol[i] = lrgba0[bobsubY[isliceoff0+i]] + lrgba1[bobsubY[isliceoff1+i]];
					}
				}
				  glTexImage2D(DHP_TEX,0,GL_RGBA,nxsb2,nzsb2,iTexBndry,
					  GL_RGBA,GL_UNSIGNED_BYTE,sol);
			};
			glBegin(GL_QUADS);
				glTexCoord2f(xx0, zz0);    glVertex3f(px0, yyv, pz0);
				glTexCoord2f(xx0, zz1);    glVertex3f(px0, yyv, pz1);
				glTexCoord2f(xx1, zz1);    glVertex3f(px1, yyv, pz1);
				glTexCoord2f(xx1, zz0);    glVertex3f(px1, yyv, pz0);
			glEnd();
			nslices++;
		}}     /* end of islice:for-loop and (iClipThis == 1):if-block */

		if(m_iTimersOn == 1) {
			if(m_bColorTableIsAvailable) { tsEvent.StampTime("End   Bob_RenderingXZ"); }
			else                         { tsEvent.StampTime("End   Bol_RenderingXZ"); }
		}
	  }

	 /**************************************************************
	  * Rendering direction is along Z-axis                        *
	  **************************************************************/
	  if(idir == 2) {
		if(m_iTimersOn == 1) {
			if(m_bColorTableIsAvailable) { tsEvent.StampTime("Begin Bob_RenderingXY"); }
			else                         { tsEvent.StampTime("Begin Bol_RenderingXY"); }
		}

		if(dzz < 0.) { ids0 = 0;                                             idsd =  1; }
		else	 	 { ids0 = ((nzsub-2*m_iNBoundaries) * (int)nplanes) - 1; idsd = -1; }

	    islicesize = nxsb2*nysb2;
//	    islicesize = nxsub*nysub;
		dz  = subvolsiz[2]/(float)((nzsub-2*m_iNBoundaries) * nplanes);
		dz1 = subvolsiz[2]/(float) (nzsub-2*m_iNBoundaries);
//      alpha = (float)alpha0 * ((float)sqrt(axx*axx+ayy*ayy+azz*azz)/(azz)) * (dz/((float)2.*cubesz[0]));
//		alpha = (float)alpha0 * (dz/((float)2.*cubesz[0]*azz));
		alpha = (float)alpha0 * (dz/((float)2.*cubesz[0]));
        glColor4d((double)1.0, (double)1.0, (double)1.0, (double)alpha);

		for(islice = 0;  islice < ((nzsub-2*m_iNBoundaries) * nplanes);  islice++) {
		zzv = subminpos[2] + dz * ((float)0.5 + (float)(ids0 + idsd*islice));

		iClipThis = iClipThis0;
		if(m_ShvrUse.ifauxclip[4] == 2  &&  zzv < m_ShvrUse.auxclip[4][3]) { iClipThis = 1; }
		if(m_ShvrUse.ifauxclip[5] == 2  &&  zzv > m_ShvrUse.auxclip[5][3]) { iClipThis = 1; }
		if(iClipThis == 0) {

			islice0  = (int)((float)0.5 + (zzv - subminpos[2]) / dz1) + m_iNBoundaries - 1;
			islice1 = islice0 + 1;
			if(islice0 <      0) { islice0 = 0; }
			if(islice1 >= nzsub) { islice1 = nzsub - 1; }

			if(m_bColorTableIsAvailable  &&  iOpac == 0) {
				if(islice0 == islice1  ||  nplanes == 1) {
					isliceoff = islicesize * islice0;
				} else {
					isliceoff0 = islicesize * islice0;
					isliceoff1 = islicesize * islice1;
					ppv0 = subminpos[2] + dz1 * ((float)0.5 + (float)(islice0-m_iNBoundaries));
					ppv1 = subminpos[2] + dz1 * ((float)0.5 + (float)(islice1-m_iNBoundaries));
					wt0 = (ppv1 - zzv) / (ppv1 - ppv0);
					if(wt0 > 0.5) { isliceoff = isliceoff0; } else { isliceoff = isliceoff1; }
				}
				glTexImage2D(DHP_TEX, 0, GL_COLOR_INDEX8_EXT,
					nxsb2,nysb2,iTexBndry, GL_COLOR_INDEX, GL_UNSIGNED_BYTE,
					&bobsubZ[isliceoff]);
			} else {
				if(islice0 == islice1  ||  nplanes == 1) {
					isliceoff = islicesize * islice0;
					if(iOpac == 0) {
						for(i=0; i<nxsb2*nysb2; i++) { sol[i] = m_lrgba[bobsubZ[isliceoff+i]]; }
					} else {
						wt0 = (float)255.99;
						for(i=0; i<nxsb2*nysb2; i++) {
							iRGB = bobsubZ[isliceoff+i];
							iALP = bobOZ[isliceoff+i];
							sol[i] =  (unsigned long)(wt0*frgba[iRGB][0])
								  + (((unsigned long)(wt0*frgba[iRGB][1]))<<8 )
								  + (((unsigned long)(wt0*frgba[iRGB][2]))<<16)
								  + (((unsigned long)(wt0*frgba[iALP][3]))<<24);
						}
					}
				} else {
					isliceoff0 = islicesize * islice0;
					isliceoff1 = islicesize * islice1;
					ppv0 = subminpos[2] + dz1 * ((float)0.5 + (float)(islice0-m_iNBoundaries));
					ppv1 = subminpos[2] + dz1 * ((float)0.5 + (float)(islice1-m_iNBoundaries));
					wt0 = (float)255.99 * (ppv1 - zzv) / (ppv1 - ppv0);

					for(i=0; i<256; i++) {
						lrgba0[i] =   (unsigned long)(wt0*frgba[i][0])
								  + (((unsigned long)(wt0*frgba[i][1]))<<8 )
								  + (((unsigned long)(wt0*frgba[i][2]))<<16)
								  + (((unsigned long)(wt0*frgba[i][3]))<<24);
						lrgba1[i] = m_lrgba[i] - lrgba0[i];
					}
					for(i=0; i<nxsb2*nysb2; i++) {
						sol[i] = lrgba0[bobsubZ[isliceoff0+i]] + lrgba1[bobsubZ[isliceoff1+i]];
					}
				}
				glTexImage2D(DHP_TEX,0,GL_RGBA,nxsb2,nysb2,iTexBndry,
					GL_RGBA,GL_UNSIGNED_BYTE,sol);
			}
			glBegin(GL_QUADS);
				glTexCoord2f(xx0, yy0);    glVertex3f(px0, py0, zzv);
				glTexCoord2f(xx0, yy1);    glVertex3f(px0, py1, zzv);
				glTexCoord2f(xx1, yy1);    glVertex3f(px1, py1, zzv);
				glTexCoord2f(xx1, yy0);    glVertex3f(px1, py0, zzv);
			glEnd();
			nslices++;
		}}     /* end of islice:for-loop and (iClipThis == 1):if-block */

//		glAccum(GL_ACCUM, alphafac);
//		glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

		if(m_iTimersOn == 1) {
			if(m_bColorTableIsAvailable) { tsEvent.StampTime("End   Bob_RenderingXY"); }
			else                         { tsEvent.StampTime("End   Bol_RenderingXY"); }
		}
	  }

	  glFlush();
	  glFinish();

      }           /* end of isoutside = 0 case: tile is inside frustum */

    }             /* end of ip loop (loop over pannels)                 */

	// Free cache buffer
	iUsed_of_iCache[iCache_of_iTile[isub]] = 0;
	if(m_iTimersOn == 1) { tsEvent.StampTime("End   jsubRenderloop"); }

    }             /* end of jsub loop                                   */

//	glAccum(GL_RETURN, 1.0);

	// Close any open files after loop over jsub
	if(m_iBufferedIO != 1) {		
//		dwGot = dhpReadFromOpenFile(bobsubY, tile_full, offset64t, (DWORD)nbytes, 0, 2);
	}

	// Past color LUT onto display if desired
	if(m_iShowColorBar > 0) {
		fZero = (float)0.0;
		fOne  = (float)1.0;
		fMOne = (float)(-1.0);

		glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        dleft     = (GLdouble)(-1.0);
        dright    = (GLdouble)( 1.0);
        dbottom   = (GLdouble)(-1.0);
        dtop      = (GLdouble)( 1.0);
        dnearclip = (GLdouble)( 0.5);
        dfarclip  = (GLdouble)( 1.5);
        glFrustum( dleft, dright, dbottom, dtop, dnearclip, dfarclip );
        gluLookAt(fZero, fZero, fZero,  fZero, fZero, fMOne,  fZero, fOne, fZero);

        glColor4d((double)1.0, (double)1.0, (double)1.0, (double)1.0);

		xx0 =  (float)1.8;
		if(m_iNTicks > 0) {
			yy0  = (float)(-1.6);
			yy1  = (float)(-1.9);
			eps  =  (float)0.01;
			eps1 =  (float)0.05;
			jlutmax = 16;
		} else {
			yy0  = (float)(-1.8);
			yy1  = (float)(-1.9);
			eps  =  (float)0.01;
			eps1 =  (float)0.01;
			jlutmax = 34;
		}
		i1 = (int)((float)128 * eps1/(xx0+eps1));
		i2 = 258 - i1;
//		if(m_iNTicks > 0) { eps1 = 0.1; }
		// background
/*
		for(i=0; i<258; i++) {
		for(j=0; j<34; j++) {
			  sol[i+258*j] = (((unsigned long)((float)100.00))<<24);

		}}
*/
		for(i=0; i<256; i++) {
		for(j=0; j<32; j++) {
			  sol[i+256*j] = (((unsigned long)((float)100.00))<<24);

		}}


		zzv   = (float)(-1.0);
		xxv0  = -xx0-eps1;
		xxv1  =  xx0+eps1;
		yyv0  =  yy0+eps;
		yyv1  =  yy1-eps;
/*
		fSMid = (float)0.5 * ((float)dnearclip + (float)dfarclip);
		for(i=0; i<3; i++) {
//			pan_lefpar[0][i] = pan_eye[0][i] + pan_dlef[0]*pan_torq[0][i];
//			pan_ritpar[0][i] = pan_eye[0][i] + pan_drit[0]*pan_torq[0][i];
//			pan_toppar[0][i] = pan_eye[0][i] + pan_dtop[0]*pan_up[0][i];
//			pan_botpar[0][i] = pan_eye[0][i] + pan_dbot[0]*pan_up[0][i];

			fLowLef[i] = fSMid*(pan_eye[0][i] + (0.9*pan_dlef[0]+0.1*pan_drit[0])*pan_torq[0][i]
											  + (0.9*pan_dbot[0]+0.1*pan_dtop[0])*  pan_up[0][i] );

			fLowRit[i] = fSMid*(pan_eye[0][i] + (0.1*pan_dlef[0]+0.9*pan_drit[0])*pan_torq[0][i]
											  + (0.9*pan_dbot[0]+0.1*pan_dtop[0])*  pan_up[0][i] );

			fTopLef[i] = fSMid*(pan_eye[0][i] + (0.9*pan_dlef[0]+0.1*pan_drit[0])*pan_torq[0][i]
											  + (0.8*pan_dbot[0]+0.2*pan_dtop[0])*  pan_up[0][i] );

			fTopRit[i] = fSMid*(pan_eye[0][i] + (0.1*pan_dlef[0]+0.9*pan_drit[0])*pan_torq[0][i]
											  + (0.8*pan_dbot[0]+0.2*pan_dtop[0])*  pan_up[0][i] );
		}
*/
//		glTexImage2D(DHP_TEX,0,GL_RGBA,258,34,1,GL_RGBA,GL_UNSIGNED_BYTE,sol);
		glTexImage2D(DHP_TEX,0,GL_RGBA,256,32,0,GL_RGBA,GL_UNSIGNED_BYTE,sol);

		glBegin(GL_QUADS);
//				glTexCoord2f(fZero, fZero);    glVertex3f(fLowLef[0], fLowLef[1], fLowLef[2]);
//				glTexCoord2f(fZero, fOne );    glVertex3f(fTopLef[0], fTopLef[1], fTopLef[2]);
//				glTexCoord2f(fOne , fOne );    glVertex3f(fTopRit[0], fTopRit[1], fTopRit[2]);
//				glTexCoord2f(fOne , fZero);    glVertex3f(fLowRit[0], fLowRit[1], fLowRit[2]);

				glTexCoord2f(fZero, fZero);    glVertex3f(xxv0, yyv0, zzv);
				glTexCoord2f(fOne , fZero);    glVertex3f(xxv1, yyv0, zzv);
				glTexCoord2f(fOne , fOne );    glVertex3f(xxv1, yyv1, zzv);
				glTexCoord2f(fZero, fOne );    glVertex3f(xxv0, yyv1, zzv);
		glEnd();

		// Color LUT
//		for(i=0; i<258; i++) { k = (i-1);  if(k < 0) { k = 0; }  if(k > 255) { k = 255; }
//		for(j=0; j<34; j++) {
		for(i=0; i<256; i++) { k = i;
		for(j=0; j<32; j++) {
			if(j < jlutmax) {
//				sol[i+258*j] =  (unsigned long)((float)255.99 * frgba[k][0])
				sol[i+256*j] =  (unsigned long)((float)255.99 * frgba[k][0])
							+ (((unsigned long)((float)255.99 * frgba[k][1]))<<8 )
							+ (((unsigned long)((float)255.99 * frgba[k][2]))<<16)
							+ (((unsigned long)((float)255.99              ))<<24);
			} else {
//				sol[i+258*j] = (unsigned long)0;
				sol[i+256*j] = (unsigned long)0;
			}
		}}
		if(m_iNTicks > 0) {
			for(j=jlutmax; j<24; j++) {
			for(itick=0; itick<m_iNTicks; itick++) {
//				sol[1+m_iTicks[itick]+258*j] =		(unsigned long)((float)255.99)
				sol[m_iTicks[itick]+256*j] =		(unsigned long)((float)255.99)
												+ (((unsigned long)((float)255.99))<<8 )
												+ (((unsigned long)((float)255.99))<<16)
												+ (((unsigned long)((float)255.99))<<24);
			}}
		}
		xxv0  = -xx0;
		xxv1  =  xx0;
		yyv0  =  yy0;
		yyv1  =  yy1;

//		glTexImage2D(DHP_TEX,0,GL_RGBA,258,34,1,GL_RGBA,GL_UNSIGNED_BYTE,sol);
		glTexImage2D(DHP_TEX,0,GL_RGBA,256,32,0,GL_RGBA,GL_UNSIGNED_BYTE,sol);
		glBegin(GL_QUADS);
				glTexCoord2f(fZero, fZero);    glVertex3f(xxv0, yyv0, zzv);
				glTexCoord2f(fOne , fZero);    glVertex3f(xxv1, yyv0, zzv);
				glTexCoord2f(fOne , fOne );    glVertex3f(xxv1, yyv1, zzv);
				glTexCoord2f(fZero, fOne );    glVertex3f(xxv0, yyv1, zzv);
		glEnd();

	    glFlush();
	    glFinish();

		if(m_iNTicks > 0) {
			if(iFirst_Test == 1) {
//				SelectObject (hDC, GetStockObject (SYSTEM_FONT)); 
				SelectObject (hDC, CreateFont(0, 0, 0, 0, FW_DONTCARE,
									 FALSE, FALSE, FALSE,
									 ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
									 DEFAULT_QUALITY, FF_ROMAN, NULL) );
				wglUseFontBitmaps (hDC, 0, 255, 1000); 
				glListBase (1000);
				iFirst_Test = 0;
			}

			glMatrixMode(GL_TEXTURE);
			glDisable(DHP_TEX);
			glDisable(GL_BLEND);

			for(itick=0; itick<m_iNTicks; itick++) {
				xx1 = -xx0 - eps1 + (float)2. * xx0 * (float)m_iTicks[itick]/(float)256.;
				glRasterPos3f(xx1, yy1, -1.);
				glCallLists (strlen(m_sTicks[itick]), GL_UNSIGNED_BYTE, m_sTicks[itick]);
			}

			glFlush();
			glFinish();
			glEnable(GL_BLEND);
			glEnable(DHP_TEX);
		}

	    glFlush();
	    glFinish();
	}

/************************************************************************
 * Plot Statistics                                                      *
 ************************************************************************/
	if(strcmp(m_sStats, "cross") == 0) { 
		// Set Background of entire frame to white, reset color of clear color to black 
		glClearColor(1.0, 1.0, 1.0, 1.0);
		glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
		glClearColor(0.0, 0.0, 0.0, 1.0);

		// Set the viewing geometry
		fZero = (float)0.0;
		fOne  = (float)1.0;
		fMOne = (float)(-1.0);
		glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        dleft     = (GLdouble)(-1.0);
        dright    = (GLdouble)( 1.0);
        dbottom   = (GLdouble)(-1.0);
        dtop      = (GLdouble)( 1.0);
        dnearclip = (GLdouble)( 0.5);
        dfarclip  = (GLdouble)( 1.5);
        glFrustum( dleft, dright, dbottom, dtop, dnearclip, dfarclip );
        gluLookAt(fZero, fZero, fZero,  fZero, fZero, fMOne,  fZero, fOne, fZero);

		// Plot labels 1st, just because ...
		xxv0  = (float)(-1.68);
		xxv1  = (float)( 1.99);
		yyv0  = (float)(-1.78);
		yyv1  = (float)( 1.90);
		zzv   = (float)(-1.00);

		// Generate Text font
		glColor3d(0.0, 0.0, 0.0);
		if(iFirst_Test == 1) {
			SelectObject (hDC, CreateFont(0, 0, 0, 0, FW_DONTCARE,
						  FALSE, FALSE, FALSE,
						  ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
						  DEFAULT_QUALITY, FF_ROMAN, NULL) );
			wglUseFontBitmaps (hDC, 0, 255, 1000); 
			glListBase (1000);
			iFirst_Test = 0;
		}
		// DO the labels
		glMatrixMode(GL_TEXTURE);
		glDisable(DHP_TEX);
		glDisable(GL_BLEND);

		for(itick=0; itick<m_iXNTicks; itick++) {
			xx1 = xxv0 - (float)0.07 + (xxv1-xxv0) * (float)m_iXTicks[itick]/(float)256.;
			glRasterPos3f(xx1, (float)(-1.9), -1.);
			glCallLists (strlen(m_sXTicks[itick]), GL_UNSIGNED_BYTE, m_sXTicks[itick]);
		}

		for(itick=0; itick<m_iYNTicks; itick++) {
			yy1 = yyv0 - (float)0.03 + (yyv1-yyv0) * (float)m_iYTicks[itick]/(float)256;
			glRasterPos3f((float)(-1.99), yy1, -1.);
			glCallLists (strlen(m_sYTicks[itick]), GL_UNSIGNED_BYTE, m_sYTicks[itick]);
		}
		glEnable(GL_BLEND);
		glEnable(DHP_TEX);

		glFlush();
		glFinish();

		// Generate RGBA LUT for counts
//		fpp = fopen("DEBUG_CROSS.txt", "w");
//		fprintf(fpp, "IRGB = %3d    %f  %f  %f\n", ind[0], red[0], grn[0], blu[0]);
		for(j=1; j<10; j++) {
//			fprintf(fpp, "IRGB = %3d    %f  %f  %f\n", ind[j], red[j], grn[j], blu[j]);
			i0 = ind[j-1];
			i1 = ind[j];
			for(i=i0; i<i1; i++) {
				fRed = (float)red[j-1] + ((float)(red[j]-red[j-1])/(float)(i1-i0)) * (float)(i-i0);
				fGrn = (float)grn[j-1] + ((float)(grn[j]-grn[j-1])/(float)(i1-i0)) * (float)(i-i0);
				fBlu = (float)blu[j-1] + ((float)(blu[j]-blu[j-1])/(float)(i1-i0)) * (float)(i-i0);
				lCrossLUT[i] =	 (unsigned long)(         fRed)
							 + (((unsigned long)(         fGrn))<<8 )
							 + (((unsigned long)(         fBlu))<<16)
							 + (((unsigned long)((float)255.99))<<24);
			}
		}

		// Map counts into RGB display array
		iMaxCounts = 0;
		for(i=0; i<256*256; i++) {
			if(iCross[i] < 1) { j = 0; }
			else              { j = 1 + (int)(32.0*log((double)iCross[i])/log(10.0)); }
			if(j > 255)       { j = 255; }
			sol[i] = lCrossLUT[ j ];

//			if(iCross[i] > 0  &&  iMaxCounts < 20  &&  i > 256) {
//				fprintf(fpp, "%3d  %3d   iCross=%5d   sol=%ld\n", i%256, i/256, iCross[i], sol[i]);
//				iMaxCounts++;
//			}
		}

//		fprintf(fpp, "m_sTileO = %s\n",m_sTileO);
//		fclose(fpp);

		// Apply Bounding Box and Tick Marks
		// Assume that X and Y ticks are already set
		lBlack =		(unsigned long)((float)  0.00)
					+ (((unsigned long)((float)  0.00))<<8 )
					+ (((unsigned long)((float)  0.00))<<16)
					+ (((unsigned long)((float)255.99))<<24);
		for(i=0; i<256; i++) {
			// Black bounding box
			if(i < 1  ||  i > 254) { for(j=0; j<512; j++) {sol[i+256*j] = lBlack; }}
			sol[i+256*  0] = lBlack;
			sol[i+256*255] = lBlack;

			// Black Horzontal Ticks
			if(i < 8) {
				for(itick=0; itick<m_iYNTicks; itick++) {
					sol[i+256*m_iYTicks[itick]] = lBlack;
				}
			}
		}
		// Black Vertical Ticks
		for(itick=0; itick<m_iXNTicks; itick++) {
			for(j=0; j<8; j++) { sol[m_iXTicks[itick]+256*j] = lBlack; }
		}

		// Plot It

		// Plot the RGBA map
        glColor4d((double)1.0, (double)1.0, (double)1.0, (double)1.0);
		glTexImage2D(DHP_TEX,0,GL_RGBA,256,256,0,GL_RGBA,GL_UNSIGNED_BYTE,sol);
		glBegin(GL_QUADS);
				glTexCoord2f(fZero, fZero);    glVertex3f(xxv0, yyv0, zzv);
				glTexCoord2f(fOne , fZero);    glVertex3f(xxv1, yyv0, zzv);
				glTexCoord2f(fOne , fOne );    glVertex3f(xxv1, yyv1, zzv);
				glTexCoord2f(fZero, fOne );    glVertex3f(xxv0, yyv1, zzv);
		glEnd();

		// Draw Labels 
/*
		plf = (PLOGFONT) LocalAlloc(LPTR, sizeof(LOGFONT)); 
		lstrcpy(plf->lfFaceName, "Arial"); 
		plf->lfWeight = FW_NORMAL; 
		GetClientRect(hWnd, &rc); 
 		SetBkMode(hDC, TRANSPARENT); 
 		plf->lfEscapement = 900;
		hfnt = CreateFontIndirect(plf);
		SelectObject(hDC, hfnt);
		TextOut(hDC, rc.right / 100, rc.bottom, m_sXTicks[0], lstrlen(m_sXTicks[0]));
		DeleteObject(hfnt);
 		SetBkMode(hDC, OPAQUE); 
 		LocalFree((LOCALHANDLE) plf); 
*/
		glFlush();
		glFinish();

		// Make sure that tick marks will be current the next time
		m_iXNTicks = 0;
		m_iYNTicks = 0;
	}

	if(strcmp(m_sStats, "pdf") == 0) {

		// Set colors as RGBA longs
		lBlack =		(unsigned long)((float)  0.00)
					+ (((unsigned long)((float)  0.00))<<8 )
					+ (((unsigned long)((float)  0.00))<<16)
					+ (((unsigned long)((float)255.99))<<24);


		lWhite =		(unsigned long)((float)255.99)
					+ (((unsigned long)((float)255.99))<<8 )
					+ (((unsigned long)((float)255.99))<<16)
					+ (((unsigned long)((float)255.99))<<24);

		lGreen =		(unsigned long)((float)  0.00)
					+ (((unsigned long)((float)160.00))<<8 )
					+ (((unsigned long)((float)  0.00))<<16)
					+ (((unsigned long)((float)255.99))<<24);

		// Get max counts
		iMaxCounts = 0;
		iTotalCounts = 0;
		for(i=0; i<256; i++) {
			iTotalCounts += iPDF[i];
			if(iMaxCounts < (int)iPDF[i]) { iMaxCounts = iPDF[i]; }
		}

		// SET X and Y ticks
		if(m_iNTicks > 0) {
			m_iXNTicks = m_iNTicks;
			for(itick=0; itick<m_iXNTicks; itick++) {
				m_iXTicks[itick] = m_iTicks[itick];
				sprintf(m_sXTicks[itick], "%s", m_sTicks[itick]);
			}
		} else {
			m_iXNTicks = 5;
			for(itick=0; itick<m_iXNTicks; itick++) {
				m_iXTicks[itick] = 64*itick;
				sprintf(m_sXTicks[itick], "%d", 64*itick);
			}
		}

		sscanf(m_sXTicks[           0], "%f", &xPDFmin);
		sscanf(m_sXTicks[m_iXNTicks-1], "%f", &xPDFmax);
		fDelX = (double)(xPDFmax - xPDFmin) / (double)(m_iXTicks[m_iXNTicks-1]-m_iXTicks[0]);


		if(fDelX < 0.) { fDelX = -fDelX; }
		fPDFfactor = 1.0 / (fDelX * (double)iTotalCounts);
		i10PDFmax = (int)(100.9 + log((double)iMaxCounts * fPDFfactor) / log(10.)) - 100;
		f10PDFmax = (float)i10PDFmax;

//fpp = fopen("DEBUG_PDF.txt", "w");
//fprintf(fpp, "xPDFmax=%f xPDFmin=%f fDelX=%f\n", xPDFmax, xPDFmin, fDelX);
//fprintf(fpp, "i10PDFmax=%d  Log_10(iMaxC*fPDFfac)=%f\n", i10PDFmax,
//		(float)(log( (double)iMaxCounts*fPDFfactor)/log(10.) )  );
//fclose(fpp);

		m_iYNTicks = 5;
		for(itick=0; itick<m_iYNTicks; itick++) {
			m_iYTicks[itick] = 100 + 100 * itick;
			sprintf(m_sYTicks[itick], "%d", itick - 4 + i10PDFmax);
		}

		// Generate pdf as a raster
		for(i=0; i<256; i++) {
			// White background
			for(j=0; j<512; j++) { sol[i+256*j] = lWhite; }

			// Green PDF
			if(iPDF[i] > 0) {
				jMax = (int)(m_iYTicks[m_iYNTicks-1] + 100.0*(log((double)iPDF[i]*fPDFfactor)/log(10.) - f10PDFmax));
			} else {
				jMax = 0;
			}
			for(j=0; j<jMax; j++) { sol[i+256*j] = lGreen; }

			// Black bounding box
			if(i < 1  ||  i > 254) { for(j=0; j<512; j++) {sol[i+256*j] = lBlack; }}
			sol[i+256*  0] = lBlack;
			sol[i+256*  1] = lBlack;
			sol[i+256*510] = lBlack;
			sol[i+256*511] = lBlack;

			// Black Horzontal Ticks
			if(i < 8) {
				for(itick=0; itick<m_iYNTicks; itick++) {
					sol[i+256*m_iYTicks[itick]] = lBlack;
				}
			}
		}
		// Black Vertical Ticks
		for(itick=0; itick<m_iXNTicks; itick++) {
			for(j=0; j<16; j++) { sol[m_iXTicks[itick]+256*j] = lBlack; }
		}


		fZero = (float)0.0;
		fOne  = (float)1.0;
		fMOne = (float)(-1.0);

		// Set the viewing geometry
		glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        dleft     = (GLdouble)(-1.0);
        dright    = (GLdouble)( 1.0);
        dbottom   = (GLdouble)(-1.0);
        dtop      = (GLdouble)( 1.0);
        dnearclip = (GLdouble)( 0.5);
        dfarclip  = (GLdouble)( 1.5);
        glFrustum( dleft, dright, dbottom, dtop, dnearclip, dfarclip );
        gluLookAt(fZero, fZero, fZero,  fZero, fZero, fMOne,  fZero, fOne, fZero);

		xxv0  = (float)(-1.78);
		xxv1  = (float)( 1.90);
		yyv0  = (float)(-1.78);
		yyv1  = (float)( 1.90);
		zzv   = (float)(-1.00);

		//====================================================================================
		// Plot it
		// Set Background of entire frame to white, reset color of clear color to black 
		glClearColor(1.0, 1.0, 1.0, 1.0);
		glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
		glClearColor(0.0, 0.0, 0.0, 1.0);

		// Lay down the RGBA map
        glColor4d((double)1.0, (double)1.0, (double)1.0, (double)1.0);
		glTexImage2D(DHP_TEX,0,GL_RGBA,256,512,0,GL_RGBA,GL_UNSIGNED_BYTE,sol);
		glBegin(GL_QUADS);
				glTexCoord2f(fZero, fZero);    glVertex3f(xxv0, yyv0, zzv);
				glTexCoord2f(fOne , fZero);    glVertex3f(xxv1, yyv0, zzv);
				glTexCoord2f(fOne , fOne );    glVertex3f(xxv1, yyv1, zzv);
				glTexCoord2f(fZero, fOne );    glVertex3f(xxv0, yyv1, zzv);
		glEnd();

		// Generate Text font
		glColor3d(0.0, 0.0, 0.0);

		if(iFirst_Test == 1) {
			SelectObject (hDC, CreateFont(0, 0, 0, 0, FW_DONTCARE,
						  FALSE, FALSE, FALSE,
						  ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
						  DEFAULT_QUALITY, FF_ROMAN, NULL) );
			wglUseFontBitmaps (hDC, 0, 255, 1000); 
			glListBase (1000);

//=============================================================================
//	plf = (PLOGFONT) LocalAlloc(LPTR, sizeof(LOGFONT)); 
//	lstrcpy(plf->lfFaceName, "Arial"); 
//	plf->lfWeight = FW_NORMAL;
//	plf->lfEscapement = 900; 
//	hfnt = CreateFontIndirect(plf); 
//	SelectObject(hDC, hfnt); 

//			SelectObject (hDC, CreateFont(100, 100, 700, 700, FW_DONTCARE,
//						  FALSE, FALSE, FALSE,
//						  ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
//						  DEFAULT_QUALITY, FF_ROMAN, NULL) );
//			wglUseFontBitmaps (hDC, 0, 255, 1500); 
//			glListBase (1500);
//=============================================================================
			iFirst_Test = 0;
		}


		// DO the labels
		glMatrixMode(GL_TEXTURE);
		glDisable(DHP_TEX);
		glDisable(GL_BLEND);

		for(itick=0; itick<m_iXNTicks; itick++) {
			xx1 = xxv0 - (float)0.07 + (xxv1-xxv0) * (float)m_iXTicks[itick]/(float)256.;
			glRasterPos3f(xx1, (float)(-1.9), -1.);
			glCallLists (strlen(m_sXTicks[itick]), GL_UNSIGNED_BYTE, m_sXTicks[itick]);
		}

		for(itick=0; itick<m_iYNTicks; itick++) {
			yy1 = yyv0 - (float)0.03 + (yyv1-yyv0) * (float)m_iYTicks[itick]/(float)512;
			glRasterPos3f((float)(-1.9), yy1, -1.);
			glCallLists (strlen(m_sYTicks[itick]), GL_UNSIGNED_BYTE, m_sYTicks[itick]);
		}
		sprintf(m_sYTicks[m_iYNTicks], "Log PDF");
		yy1 = yyv0 + (yyv1-yyv0) * (float)470/(float)512.;
		glRasterPos3f(xxv0+(float)(0.05), yy1, -1.);
		glCallLists (strlen(m_sYTicks[m_iYNTicks]), GL_UNSIGNED_BYTE, m_sYTicks[m_iYNTicks]);

		glEnable(GL_BLEND);
		glEnable(DHP_TEX);

		glFlush();
		glFinish();

		// Make sure that tick marks will be current the next time
		m_iXNTicks = 0;
		m_iYNTicks = 0;
	}                   // END PDF block

	sprintf(m_sStats, "none");      // Do not get stuck in cross-correlation mode

	if(m_iNRendBuffs == 2) {
		SwapBuffers(hDC);
	    glFlush();
	    glFinish();
	}
	m_iRender = 0;        // will stop rendering upon return
/*
	glGetIntegerv(GL_ACCUM_RED_BITS,   &fRBits);
	glGetIntegerv(GL_ACCUM_GREEN_BITS, &fGBits);
	glGetIntegerv(GL_ACCUM_BLUE_BITS,  &fBBits);
	glGetIntegerv(GL_ACCUM_ALPHA_BITS, &fABits);

	fpp = fopen("AccumBits.txt", "w");
	fprintf(fpp, "Number of Red    bits = %d\n", (int)fRBits);
	fprintf(fpp, "Number of Green  bits = %d\n", (int)fGBits);
	fprintf(fpp, "Number of Blue   bits = %d\n", (int)fBBits);
	fprintf(fpp, "Number of Alpha  bits = %d\n", (int)fABits);
	fclose(fpp);
*/
	if(m_iTimersOn == 1) {
		tsEvent.StampTime("End   Rendering");
		tsEvent.WriteToFile();
	}

	return;
}

void crossprod(float vec1[3], float vec2[3], float vec3[3]) {
    vec3[0] = vec1[1] * vec2[2]  -  vec1[2] * vec2[1];
    vec3[1] = vec1[2] * vec2[0]  -  vec1[0] * vec2[2];
    vec3[2] = vec1[0] * vec2[1]  -  vec1[1] * vec2[0];
}

/***************************************************************************
 *  The integer function strdif(str1, str2) takes two string arguments.    *
 *  If the strings are identical (ending in '\0') strdif returns 0.        *
 *  If there are any differences, strdif returns 1.                        *
 ***************************************************************************/
int strdif(char* str1, char* str2)
{
    int i, idiff;
    i = 0;
    while(str1[i] == str2[i] && str1[i] != '\0') { i++; }
    idiff = 1;
    if(str1[i] == str2[i]) { idiff = 0; }
    return(idiff);
}
void setstr(char* str1, char* str2)
{
    int i;
    i = 0;
    while(str2[i] != '\0') { str1[i] = str2[i]; i++; }
    str1[i] = '\0';
}


void Chvr_work::init()
{
  int i;

    glClearColor(0.0, 0.0, 0.0, 1.0);

    glEnable(GL_TEXTURE_2D);
    glEnable(GL_COLOR_INDEX8_EXT);
//    glEnable(GL_COLOR_INDEX);
    glEnable(GL_BLEND);

    glDisable(GL_DEPTH_TEST);
    glDisable(GL_DITHER);
    glDisable(GL_FOG);
    glDisable(GL_LIGHTING);
    glDisable(GL_STENCIL_TEST);
    glDisable(GL_TEXTURE_1D);
    glDisable(GL_LOGIC_OP);

    /*    glBlendFunc(GL_SRC_ALPHA, GL_ONE); */
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

    /* Similar to defining a 2D texture, but note the setting of the */
    /* wrap parameter for the R coordinate.  Also, for 3D textures   */
    /* you probably won't need mipmaps, hence the linear min filter. */

    /* glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST ); */
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	/* DETERMINE if paletted textures of the needed type and width are available */
	if(wglGetProcAddress("glColorTableEXT") != NULL) {
		glColorTableEXT = (PFNGLCOLORTABLEPROC) wglGetProcAddress("glColorTableEXT");
/*
		if(wglGetProcAddress("glGetColorTableParameterivEXT") != NULL) {
			glGetColorTableParameterivEXT =
			(PFNGLGETCOLORTABLEPARAMETERIVEXTPROC) wglGetProcAddress("glGetColorTableParameterivEXT");
			glColorTableEXT(GL_PROXY_TEXTURE_2D, GL_RGBA,256,GL_RGBA, GL_FLOAT,frgba);
			i = 0;
			glGetColorTableParameterivEXT(GL_PROXY_TEXTURE_2D, GL_COLOR_TABLE_WIDTH_EXT, &i);
			if(i == 256) { m_bColorTableIsAvailable = TRUE; }
		}
*/
	}

	/* CREATE and SET named textures */
    glGenTextures(256, texname);
    for(i=0; i<256; i++) {
        glBindTexture(DHP_TEX, texname[i]);
        glTexParameteri(DHP_TEX, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(DHP_TEX, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(DHP_TEX, GL_TEXTURE_WRAP_S, GL_CLAMP);
        glTexParameteri(DHP_TEX, GL_TEXTURE_WRAP_T, GL_CLAMP);
//        glTexParameteri(DHP_TEX, GL_TEXTURE_WRAP_S, GL_REPEAT);
//        glTexParameteri(DHP_TEX, GL_TEXTURE_WRAP_T, GL_REPEAT);
    }
    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glFlush();
	if(m_iNRendBuffs == 2) {  SwapBuffers(hDC); }

	iFirst_Test = 1;

}
/*
bool Chvr_work::DoRender()
{
	bool bRet;
	bRet = FALSE;
	if(m_iRender == 1) bRet = TRUE;
	return bRet;

}
*/
STDMETHODIMP Chvr_work::SetDoRender()
{
	// Only one thread should be running draw_scene at a time
//	int ii;
//	ii = InterlockedIncrement(&iNRendering);
//	while(ii > 1) {
//		ii = InterlockedDecrement(&iNRendering);
//		_sleep(10);
//		ii = InterlockedIncrement(&iNRendering);
//	}
//	m_iRender = 1;

	// Copy THIS objects versions of inputs to global copy for rendering thread to use
	// See CopyHvrInputs
	memcpy(&global_ShvrIn.iNewHV, &m_ShvrIn.iNewHV, sizeof(HvrInput));

	iDoThis = 1;
	if(m_iBlockOnRender == 1) { while(iDoThis == 1) { _sleep(10); } }
//	draw_scene();

//	ii = InterlockedDecrement(&iNRendering);

	return S_OK;
}

STDMETHODIMP Chvr_work::Set_HVname(unsigned char *cHVnew)
{
	sprintf(m_ShvrIn.sHVname, "%s\0", cHVnew);

//	void  setstr(char*, char*);
//	setstr(hvname, cHVnew);

	return S_OK;
}


/*****************************************************************************
 *                        READ IN PANNEL INFO HERE                           *
 *---------------------------------------------------------------------------*
 * read in of configuration file goes here : configuration file gives rules  *
 * for generating [center? and (horiz, vert) offsets]                        *
 * for each pannel for each pannel given center?, eye?, up? from above       *
 *---------------------------------------------------------------------------*
 *              torq    eye     up                                           *
 *            i    0      1      2                                           *
 *  pw_normal[i]   0.     1.     0.    <-- typical example of central pannel *
 *      pw_up[i]   0.     0.     1.    <-- typical example of central pannel *
 *                                                                           *
 *---------------------------------------------------------------------------*
 * Example PW_CONFIGURATION for single pannel (1280x1024 per pannel)         *
 *       0. 1. 0.  0. 0. 1.  0. 1. 0.  0.625 0.5                             *
 *---------------------------------------------------------------------------*
 * Example PW_CONFIGURATION for flat 4 pannel (1600x1200 per pannel)         *
 *       0. 1. 0.  0. 0. 1.  -0.6667 1.0 -0.5  0.6667 0.5                    *
 *       0. 1. 0.  0. 0. 1.   0.6667 1.0 -0.5  0.6667 0.5                    *
 *       0. 1. 0.  0. 0. 1.  -0.6667 1.0  0.5  0.6667 0.5                    *
 *       0. 1. 0.  0. 0. 1.   0.6667 1.0  0.5  0.6667 0.5                    *
 *****************************************************************************/

STDMETHODIMP Chvr_work::AddFrustum(float *center, float *up, float *normal, int iWidth, int iHeight, float fPanWidth)
{
	int i;
	float dot;

	m_ShvrIn.iNewFrustum = 1;
	m_ShvrIn.iWidth  = iWidth;
	m_ShvrIn.iHeight = iHeight;
    m_ShvrIn.fAspectRatio = (float)iHeight / (float)iWidth;

	for(i=0; i<3; i++) { m_ShvrIn.fPanCent[i] = center[i]; }
	for(i=0; i<3; i++) { m_ShvrIn.fPanNorm[i] = normal[i]; }
	for(i=0; i<3; i++) { m_ShvrIn.fPanUp[i]   = up[i]; }

	m_ShvrIn.fHalfWidth = (float)0.5 * fPanWidth;
	m_ShvrIn.fHalfHeight = m_ShvrIn.fAspectRatio * m_ShvrIn.fHalfWidth;

	dot = center[0]*normal[0] + center[1]*normal[1] + center[2]*normal[2];
	if(dot < 0.) { dot = -dot; }
	m_ShvrIn.fFovAngle = (float)2.0 * (float)atan(m_ShvrIn.fHalfHeight/dot) / Pid180;

	npannel = 1;

	return S_OK;
}

STDMETHODIMP Chvr_work::SetFrustum(float *center, float *up, float *normal, int iWidth, int iHeight, float fPanWidth)
{
	int i;
	float dot;

	m_ShvrIn.iNewFrustum = 1;
	m_ShvrIn.iWidth  = iWidth;
	m_ShvrIn.iHeight = iHeight;
    m_ShvrIn.fAspectRatio = (float)iHeight / (float)iWidth;

	for(i=0; i<3; i++) { m_ShvrIn.fPanCent[i] = center[i]; }
	for(i=0; i<3; i++) { m_ShvrIn.fPanNorm[i] = normal[i]; }
	for(i=0; i<3; i++) { m_ShvrIn.fPanUp[i]   = up[i]; }

	m_ShvrIn.fHalfWidth = (float)0.5 * fPanWidth;
	m_ShvrIn.fHalfHeight = m_ShvrIn.fAspectRatio * m_ShvrIn.fHalfWidth;

	dot = center[0]*normal[0] + center[1]*normal[1] + center[2]*normal[2];
	if(dot < 0.) { dot = -dot; }
	m_ShvrIn.fFovAngle = (float)2.0 * (float)atan(m_ShvrIn.fHalfHeight/dot) / Pid180;

	npannel = 1;

	return S_OK;
}

void Chvr_work::SetDefaultFrustum()
{

	m_ShvrIn.iNewFrustum = 1;
	m_ShvrIn.iWidth  = 1024;
	m_ShvrIn.iHeight = 1024;
    m_ShvrIn.fAspectRatio = (float)m_ShvrIn.iWidth / (float)m_ShvrIn.iHeight;

    m_ShvrIn.fPanNorm[0] = 0.;    /* default single pannel settings */
    m_ShvrIn.fPanNorm[1] = 1.;
    m_ShvrIn.fPanNorm[2] = 0.;
    m_ShvrIn.fPanUp[0]   = 0.;
    m_ShvrIn.fPanUp[1]   = 0.;
    m_ShvrIn.fPanUp[2]   = 1.;
	m_ShvrIn.fPanCent[0] = 0.;
    m_ShvrIn.fPanCent[1] = 1.;
	m_ShvrIn.fPanCent[2] = 0.;
	
	m_ShvrIn.fFovAngle  = 34.;
    m_ShvrIn.fHalfHeight = (float)tan( Pid180* m_ShvrIn.fFovAngle * 0.5 );
    m_ShvrIn.fHalfWidth  = m_ShvrIn.fHalfHeight / m_ShvrIn.fAspectRatio;

	npannel = 1;
}
/*
	struct HvrInput {    // This structure contains all of the input needed for the hvr code
		int iNewHV;
		char sHVname[100];

		int iNewFrustum;
		int iWidth;
		int iHeight;
		float fPanCent[3];
		float fPanNorm[3];
		float fPanUp[3];
		float fHalfWidth;
		float fHalfHeight;
		float fFovAngle;
		float fAspectRatio;

		int iNewLUT;
		char sCmapname[100];
		char sAmapname[100];

		int iNewView;
		float fEye[3];
		float fCent[3];
		float fUp[3];
		float fNearClip;
		float fFarrClip;
	};
	HvrInput m_ShvrIn;     // GUI writes input here
	HvrInput m_ShvrUse;    // Rendering code uses ShvrUse, after copying info from ShvrIn
*/

void Chvr_work::SetDefaultView()
{
	m_ShvrIn.iNewView = 1;
	m_ShvrIn.fCent[0] = 0.;
	m_ShvrIn.fCent[1] = 0.;
	m_ShvrIn.fCent[2] = 0.;
	m_ShvrIn.fEye[0] = 0.0;
	m_ShvrIn.fEye[1] = 0.0;
	m_ShvrIn.fEye[2] = 10.0;
	m_ShvrIn.fUp[0] = 0.;
	m_ShvrIn.fUp[1] = 1.;
	m_ShvrIn.fUp[2] = 0.;
	m_ShvrIn.fNearClip = 0.5;
	m_ShvrIn.fFarrClip = 20.5;
}

void Chvr_work::CopyHvrInputs()
{
	// Copy from GUI input struct to hvr use struct
	// Set all iNew flags to the status of NOT new.
    // BEGIN NEED THREAD SAFE!!
	memcpy(&m_ShvrIn.iNewHV,  &global_ShvrIn.iNewHV, sizeof(HvrInput));
	memcpy(&m_ShvrUse.iNewHV, &global_ShvrIn.iNewHV, sizeof(HvrInput));
	opacity   = global_ShvrIn.opacity;
	tolerance = global_ShvrIn.tolerance;		/* max color level error allowed with no scan down  */
	angtol    = global_ShvrIn.angtol;			/* max angle allowed with no scan down              */
	viewres   = global_ShvrIn.viewres;			/* dx_brick > farclip * vrtolerance => no scan down */
	nplanes   = global_ShvrIn.nplanes;
	maxplanes = global_ShvrIn.maxplanes;

//	memcpy(&m_ShvrUse.iNewHV, &m_ShvrIn.iNewHV, sizeof(HvrInput));
	m_ShvrIn.iNewHV      = 0;
	m_ShvrIn.iNewLUT     = 0;
	m_ShvrIn.iNewView    = 0;
	m_ShvrIn.iNewFrustum = 0;
    // END NEED THREAD SAFE!!
}

STDMETHODIMP Chvr_work::IncrementEyePoint(int delx, int dely, int iMouseMode)
{
    void crossprod(float vec1[3], float vec2[3], float vec3[3]);
	void vnormalize( float v[3] );
    float dotproduct( const float*, const float* );
	float rotfacx, rotfacy;
	float fEC[3], fHoriz[3], fVert[3], fDelUp, fEN[3], fNewEN[3], fNormEN2, fNormNewEN2;
	float fNewEC[3], fNormEC2, fNormEC, fNormNewEC2;
	float fReNormFac, fNewDelUp, fDelCent, ffac;
	int i;

	for(i=0; i<3; i++) { fEC[i] = m_ShvrIn.fEye[i] - m_ShvrIn.fCent[i]; }
	vnormalize(m_ShvrIn.fUp);
	crossprod(fEC, m_ShvrIn.fUp, fHoriz);

	if(iMouseMode == 0  ||  iMouseMode == 3) {           //  Polar or Track Ball Mode
		rotfacx = (float)0.0100 * (float)delx;
		rotfacy = (float)0.0050 * (float)dely;
		vnormalize(fHoriz);
		crossprod(fHoriz, fEC, fVert);
		vnormalize(fVert);
		fNormEC2 = dotproduct(fEC, fEC);
		fNormEC  = (float)sqrt(fNormEC2);
		for(i=0; i<3; i++) { fNewEC[i] = fEC[i] + fNormEC * (rotfacx * fHoriz[i] + rotfacy * fVert[i]); }
		fNormNewEC2 = dotproduct(fNewEC, fNewEC);
		fReNormFac = (float)sqrt(fNormEC2 / fNormNewEC2);
		for(i=0; i<3; i++) { fNewEC[i] *= fReNormFac; }    // Want norms of NewEC and old EC to be equal
		for(i=0; i<3; i++) { m_ShvrIn.fEye[i] = m_ShvrIn.fCent[i] + fNewEC[i]; }

		if(iMouseMode == 3) {                             // Track Ball Mode
			crossprod(fNewEC, m_ShvrIn.fUp, fHoriz);
			crossprod(fHoriz, fNewEC, m_ShvrIn.fUp);
			vnormalize(m_ShvrIn.fUp);
		}
	}

	if(iMouseMode == 1) {                                 // Cylinder Mode
		rotfacx = (float)0.0100 * (float)delx;
		rotfacy = (float)0.0010 * (float)dely;
		fDelUp = dotproduct(m_ShvrIn.fUp, fEC);
		for(i=0; i<3; i++) { fEN[i]    = fEC[i] - fDelUp * m_ShvrIn.fUp[i]; }
		for(i=0; i<3; i++) { fNewEN[i] = fEN[i] + rotfacx * fHoriz[i]; }
		fNormEN2    = dotproduct(fEN, fEN);
		fNormNewEN2 = dotproduct(fNewEN, fNewEN);
		fReNormFac = (float)sqrt(fNormEN2 / fNormNewEN2);
		for(i=0; i<3; i++) { fNewEN[i] *= fReNormFac; }    // Want norms of NewEN and old EN to be equal
		fDelCent  = rotfacy * (float)sqrt(fNormEN2);
		fNewDelUp = fDelUp + fDelCent;
		for(i=0; i<3; i++) {
			m_ShvrIn.fEye[i]  = m_ShvrIn.fCent[i] + fNewEN[i] + fNewDelUp * m_ShvrIn.fUp[i];
			m_ShvrIn.fCent[i] = m_ShvrIn.fCent[i]             + fDelCent  * m_ShvrIn.fUp[i];
		}
	}

	if(iMouseMode == 2) {                                   // Pan Mode
		crossprod(fHoriz, fEC, fVert);
		fNormEC = (float)sqrt(dotproduct(fEC, fEC));
		vnormalize(fHoriz);
		vnormalize(fVert);

		rotfacx = (float)0.00065 * fNormEC * (float)delx;
		rotfacy = (float)0.00065 * fNormEC * (float)dely;
		ffac = (float)(1. / sqrt(dotproduct(fHoriz, fHoriz)));
		for(i=0; i<3; i++) {
			m_ShvrIn.fEye[i]  += rotfacx * fHoriz[i] + rotfacy * fVert[i];
			m_ShvrIn.fCent[i] += rotfacx * fHoriz[i] + rotfacy * fVert[i];
		}
	}


	return S_OK;
}

STDMETHODIMP Chvr_work::SetRenderQuality(int iPixErr, float fAngleError, float fSizeError)
{
	tolerance = iPixErr;		/* max color level error allowed with no scan down  */
	angtol    = fAngleError;	/* max angle allowed with no scan down              */
	viewres   = fSizeError;	    /* dx_brick > farclip * vrtolerance => no scan down */
	return S_OK;
}

STDMETHODIMP Chvr_work::StopRendering()
{
	if(iDoThis == 1) {
		iStopRendering = 1;
		iIO_Read = 2;
	}
	return S_OK;
}

STDMETHODIMP Chvr_work::SaveKey(struct Renderparams *SKey)
{
	int i, j;

	SKey->nx = Nx;
	SKey->ny = Ny;
	SKey->nz = Nz;
	SKey->opacity = opacity;
	sprintf((char *)SKey->cmapname,"%s", m_ShvrIn.sCmapname);
	sprintf((char *)SKey->amapname,"%s", m_ShvrIn.sAmapname);
	*SKey->bobname = NULL;
	sprintf((char *)SKey->hvname, "%s", m_ShvrIn.sHVname);
	SKey->isectx = 0;
	SKey->isecty = 0;
	SKey->nsectx = 1;
	SKey->nsecty = 1;
	SKey->width = m_ShvrIn.iWidth;
	SKey->height = m_ShvrIn.iHeight;
	SKey->hpos = 0;
	SKey->vpos = 0;
	SKey->moviemode = 0;
	SKey->movieframes = 1;
	sprintf((char *)SKey->imagename, "Test_Image.pix");
	sprintf((char *)SKey->moviename, "Test_Movie.pix");
	SKey->fovangle = m_ShvrIn.fFovAngle;
	SKey->eyeX = m_ShvrIn.fEye[0];
	SKey->eyeY = m_ShvrIn.fEye[1];
	SKey->eyeZ = m_ShvrIn.fEye[2];
	SKey->centerX = m_ShvrIn.fCent[0];
	SKey->centerY = m_ShvrIn.fCent[1];
	SKey->centerZ = m_ShvrIn.fCent[2];
	SKey->upX = m_ShvrIn.fUp[0];
	SKey->upY = m_ShvrIn.fUp[1];
	SKey->upZ = m_ShvrIn.fUp[2];
	SKey->eyedirX = m_ShvrIn.fCent[0] - m_ShvrIn.fEye[0];
	SKey->eyedirY = m_ShvrIn.fCent[1] - m_ShvrIn.fEye[1];
	SKey->eyedirZ = m_ShvrIn.fCent[2] - m_ShvrIn.fEye[2];
	for(i=0; i<4; i++) { SKey->viewdefined[i] = 1; }
	for(i=0; i<6; i++) { SKey->ifauxclip[i] = m_ShvrIn.ifauxclip[i]; }
	for(i=0; i<6; i++) { for(j=0; j<4; j++) { SKey->auxclip[i][j] = m_ShvrIn.auxclip[i][j]; }}
	SKey->my_near = m_ShvrIn.fNearClip;
	SKey->my_far  = m_ShvrIn.fFarrClip;
	SKey->softdistance = 0.;
	SKey->nplanes = nplanes;
	SKey->maxplanes = maxplanes;
	SKey->boxcolor = -1;
	SKey->saveslabs = 0;
	SKey->movieframe = 0;
	SKey->nfpk = 10;
	SKey->ntpf = 0;
	SKey->nrpk = 0.;
	SKey->icoltolerance = tolerance;
	SKey->angletolerance = angtol;
	SKey->vrtolerance = viewres;

	return S_OK;
}

STDMETHODIMP Chvr_work::RestorKey(struct Renderparams *SKey)
{
	int i, j;

	sprintf(m_ShvrIn.sCmapname, "%s", SKey->cmapname);
	sprintf(m_ShvrIn.sAmapname, "%s", SKey->amapname);
	sprintf(m_ShvrIn.sHVname, "%s", SKey->hvname);
	m_ShvrIn.fFovAngle = SKey->fovangle;
	m_ShvrIn.fEye[0] = SKey->eyeX;
	m_ShvrIn.fEye[1] = SKey->eyeY;
	m_ShvrIn.fEye[2] = SKey->eyeZ;
	m_ShvrIn.fCent[0] = SKey->centerX;
	m_ShvrIn.fCent[1] = SKey->centerY;
	m_ShvrIn.fCent[2] = SKey->centerZ;
	m_ShvrIn.fUp[0] = SKey->upX;
	m_ShvrIn.fUp[1] = SKey->upY;
	m_ShvrIn.fUp[2] = SKey->upZ;
	m_ShvrIn.fNearClip = SKey->my_near;
	m_ShvrIn.fFarrClip = SKey->my_far;
	for(i=0; i<6; i++) { m_ShvrIn.ifauxclip[i] = SKey->ifauxclip[i]; }
	for(i=0; i<6; i++) { for(j=0; j<4; j++) { m_ShvrIn.auxclip[i][j] = SKey->auxclip[i][j]; }}
	sprintf(m_cImageName, "%s", SKey->imagename);

	opacity = SKey->opacity;
	tolerance = SKey->icoltolerance;				/* max color level error allowed with no scan down  */
	angtol    = SKey->angletolerance;				/* max angle allowed with no scan down              */
	viewres   = SKey->vrtolerance;					/* dx_brick > farclip * vrtolerance => no scan down */
	nplanes   = SKey->nplanes;
	maxplanes = SKey->maxplanes;

	m_ShvrIn.opacity = opacity;
	m_ShvrIn.tolerance = tolerance;
	m_ShvrIn.angtol = angtol;
	m_ShvrIn.viewres = viewres;
	m_ShvrIn.nplanes = nplanes;
	m_ShvrIn.maxplanes = maxplanes;

	return S_OK;
}

STDMETHODIMP Chvr_work::ZoomEyePoint(int delx, int dely, int iMouseMode)
{
	float fMult, fEC[3];
	int i;

	fMult = (float)exp(-0.002 * (double)dely);

	for(i=0; i<3; i++) { fEC[i] = m_ShvrIn.fEye[i] - m_ShvrIn.fCent[i]; }
	for(i=0; i<3; i++) { m_ShvrIn.fEye[i] = m_ShvrIn.fCent[i] + fMult * fEC[i]; }

	if(iMouseMode == 2) {   // Pan Mode
		for(i=0; i<3; i++) { m_ShvrIn.fCent[i] = m_ShvrIn.fEye[i] - fEC[i]; }
	}

	return S_OK;
}

STDMETHODIMP Chvr_work::SetLUT(unsigned char *sAmap, unsigned char *sCmap)
{
	setstr(m_ShvrIn.sCmapname, (char *)sCmap);
	setstr(m_ShvrIn.sAmapname, (char *)sAmap);
	return S_OK;
}

STDMETHODIMP Chvr_work::CopyPixToFile(unsigned char *cFile)
{
	FILE *fdout;
	int nbytes, npix, i0, i1;
	char string[100], cOutFile[100];

//	glReadBuffer(GL_FRONT);
	if(m_iNRendBuffs == 2) { SwapBuffers(hDC); }

	nbytes = m_ShvrUse.iHeight * m_ShvrUse.iWidth * 3;

    glReadPixels( 0, 0, m_ShvrUse.iWidth, m_ShvrUse.iHeight,
		  GL_RGB, GL_UNSIGNED_BYTE, rgbpannel[0]);
	if(m_iNRendBuffs == 2) { SwapBuffers(hDC); }

	setstr(string, (char *)cFile);

	i0 = (int)string[0] - (int)'0';
	i1 = (int)string[1] - (int)'0';
	if(0 <= i0 && i0 <= 9 && 0 <= i1  &&  i1 <= 9) {
		sprintf(cOutFile, "%s_%c%c", m_cImageName, string[0], string[1]);
	} else {
		setstr(cOutFile, (char *)cFile);
	}

    fdout = fopen(cOutFile, "wb");
    if ( fdout == NULL ) { return S_OK; }

    npix = fwrite(rgbpannel[0], sizeof(char), nbytes, fdout);
//    if ( npix != nbytes ) { fclose(fdout); return (S_FAIL); }
    fclose(fdout);
/*
	fp = fopen("DebugCopyPix.txt", "w");
	fprintf(fp, "wrote %d bytes into file %s\n", nbytes, cOutFile);
	fprintf(fp, "Height = %d  Width = %d\n", m_ShvrUse.iHeight, m_ShvrUse.iWidth);
	if(npix != nbytes) { fprintf(fp, "write failed, actually wrote: %d bytes\n", npix); }
	fclose(fp);
*/

	return S_OK;
}

STDMETHODIMP Chvr_work::SetCurrentWorkingDir(unsigned char *cDir)
{
	char string[100];
//	FILE *fp;

	setstr(string, (char *)cDir);
	SetCurrentDirectory( string );
//	fp = fopen("ImHere", "w");
//	fprintf(fp, "directory = %s\n", string);
//	fclose(fp);

	return S_OK;
}

STDMETHODIMP Chvr_work::SetOpacity(float fOpacity)
{
	opacity = fOpacity;
	return S_OK;
}

STDMETHODIMP Chvr_work::SetNearFarClip(float fNear, float fFar)
{
	m_ShvrIn.fNearClip = fNear;
	m_ShvrIn.fFarrClip = fFar;

	return S_OK;
}

DWORD dhpReadFromOpenHandle(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes, int iClose)
{
	static HANDLE hFile;                        // handle to read file
	static int iIsOpen = 0;
	__int64 shift32, shift8;
	LONG dwHigh, dwLow;
	unsigned long ngot, nread;
    static OVERLAPPED	overlapped;
	BOOL fs;

	if((iClose > 0)  &&  (iIsOpen > 0)) {
		CloseHandle(hFile);
		iIsOpen = 0;
	}

	if(iClose > 1) { return 0; }

	if(iIsOpen == 0) {
		hFile = CreateFile(	lpFileName,
							GENERIC_READ, 
							FILE_SHARE_READ,
							NULL, 
							OPEN_EXISTING, 
							FILE_FLAG_NO_BUFFERING | FILE_FLAG_OVERLAPPED, 
							NULL											);

		if (hFile == INVALID_HANDLE_VALUE) { return 0; }

		iIsOpen = 1;
	}

	shift8  = (__int64)256;
	shift32 = shift8 * shift8 * shift8 * shift8;
	dwHigh  = (LONG)(offset / shift32);
	dwLow   = (LONG)(offset % shift32);
	overlapped.Offset = dwLow;
	overlapped.OffsetHigh = dwHigh;
	overlapped.hEvent = NULL;

	fs = ReadFile(hFile, lpBuffer, nbytes, &nread, &overlapped);
	GetOverlappedResult(hFile, &overlapped, &ngot, TRUE);

	return ngot;
}


DWORD dhpReadFromOpenFile(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes,
						  int itcd, int iClose)
{
	static HANDLE hFiles[8192];
	static int iFirst = 1, iIsOpen[8192];
	int i;
	HANDLE hFile;                        // handle to read file
	__int64 shift32, shift8;
	LONG dwHigh, dwLow;
	unsigned long ngot;
//	DWORD lasterror;
//	LPSTR lpsErrorMsg;

	shift8  = (__int64)256;
	shift32 = shift8 * shift8 * shift8 * shift8;
	dwHigh  = (LONG)(offset / shift32);
	dwLow   = (LONG)(offset % shift32);

	if(iFirst == 1) {
		for(i=0; i<8192; i++) { iIsOpen[i] = 0; }
		iFirst = 0;
	}

	if(iClose == 2) {
		for(i=0; i<8192; i++) {
			if(iIsOpen[i] != 0) {
				hFile = hFiles[i];
				CloseHandle(hFile);
				iIsOpen[i] = 0;
			}
		}
		return 0;
	}

	if(iIsOpen[itcd] == 0) {
		hFile = CreateFile(lpFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
							FILE_ATTRIBUTE_NORMAL, NULL);

		if(hFile == INVALID_HANDLE_VALUE) {
			//lasterror = GetLastError();
//			printf("Got an invalid file handle for file %s\n", lpFileName);
			//FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |
			//        FORMAT_MESSAGE_IGNORE_INSERTS, NULL, lasterror, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			//		lpsErrorMsg, 0, NULL);
//			printf("Tried to open file    -->%s<--\n", lpFileName);
//			printf("OK, lets try that again\n");
//			printf("Error on CreateFile was: %ld\n", lasterror); //, lpsErrorMsg);
//			exit(0);
		}

		hFiles[itcd] = hFile;
		iIsOpen[itcd] = 1;
//		printf("----Openned file: itcd = %d\n", itcd);
	}
	hFile = hFiles[itcd];
	
	SetFilePointer(hFile, dwLow, &dwHigh, FILE_BEGIN);
	ReadFile(hFile, lpBuffer, nbytes, &ngot, NULL);

	if(ngot != nbytes) {
//		lasterror = GetLastError();
//		printf("----Only read %ld out of %ld bytes\n", *ngot, nbytes, lpFileName);
//		printf("----LastError was: %ld\n", lasterror);//, lpsErrorMsg);
//		printf("----Offset = %ld\n", offset);
	}
	
	if(iClose == 1) {
		CloseHandle(hFile);
		iIsOpen[itcd] = 0;
	}

	return ngot;
}

DWORD dhpReadFromFileSync(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes)
{
	HANDLE hFile;                        // handle to read file
	__int64 shift32, shift8;
	LONG dwHigh, dwLow;
	unsigned long ngot;

	shift8  = (__int64)256;
	shift32 = shift8 * shift8 * shift8 * shift8;
	dwHigh  = (LONG)(offset / shift32);
	dwLow   = (LONG)(offset % shift32);
	hFile = CreateFile(lpFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL, NULL);
	SetFilePointer(hFile, dwLow, &dwHigh, FILE_BEGIN);
	ReadFile(hFile, lpBuffer, nbytes, &ngot, NULL);
	CloseHandle(hFile);

	return ngot;
}

DWORD dhpReadFromFileNoBuf(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes)
{
	HANDLE hFile;                        // handle to read file
	__int64 shift32, shift8;
	LONG dwHigh, dwLow;
	unsigned long ngot, nread;
    static OVERLAPPED	overlapped;
	BOOL fs;

	hFile = CreateFile(	lpFileName,
						GENERIC_READ, 
						FILE_SHARE_READ,
						NULL, 
						OPEN_EXISTING, 
						FILE_FLAG_NO_BUFFERING | FILE_FLAG_OVERLAPPED, 
						NULL											);

	if (hFile == INVALID_HANDLE_VALUE) { return 0; }

	shift8  = (__int64)256;
	shift32 = shift8 * shift8 * shift8 * shift8;
	dwHigh  = (LONG)(offset / shift32);
	dwLow   = (LONG)(offset % shift32);
	overlapped.Offset = dwLow;
	overlapped.OffsetHigh = dwHigh;
	overlapped.hEvent = NULL;

	fs = ReadFile(hFile, lpBuffer, nbytes, &nread, &overlapped);
	GetOverlappedResult(hFile, &overlapped, &ngot, TRUE);

	CloseHandle(hFile);

	return ngot;
}

int TwoBrickIO(unsigned char* bricks[2], int iNBytes, char* filename)
{
	static int				fs;
	int						iCurrentBuffer;
    static unsigned long	bytesread, bytestransferred;
    static int				bufno = -1;
    static HANDLE			fd; 
    static OVERLAPPED		overlapped;

	iCurrentBuffer = bufno;

    if (bufno >= 0)   // Finish Previous Read & Close Handle to File
    {
      GetOverlappedResult(fd, &overlapped, &bytestransferred, TRUE);
      CloseHandle(fd);
    }

    if (iNBytes > 0)  // Open Handle to File and Initiate Read
    {
		fd = CreateFile(filename,
				GENERIC_READ, 
				FILE_SHARE_READ,
				0, 
				OPEN_EXISTING, 
				FILE_FLAG_NO_BUFFERING | FILE_FLAG_OVERLAPPED, 
				0);

		if (fd == INVALID_HANDLE_VALUE)
		{
			printf("Could not open a handle to file: %s\n", filename);
			exit(E_FAIL);
		}
		bufno = (bufno + 1)%2;
		overlapped.Offset = 0;
		overlapped.OffsetHigh = 0;
		overlapped.hEvent = NULL;
		fs = ReadFile(fd, bricks[bufno], iNBytes, &bytesread, &overlapped);
    }

    return(iCurrentBuffer); 
}

STDMETHODIMP Chvr_work::Set_hWnd(int i_hWnd)
{

	m_hLocalWnd = (HWND)i_hWnd;

	return S_OK;
}

STDMETHODIMP Chvr_work::SetLUTirgb(int N, int *ind, float *red, float *grn, float *blu)
{
	int i0, i1, i, j;
	float y0, yp;

	m_iNrgbNots = N;
	for(j=0; j<N; j++) {
		m_iRGBind[j] = ind[j];
		m_fRED[j]    = red[j];
		m_fGRN[j]    = grn[j];
		m_fBLU[j]    = blu[j];
	}

	for(j=1; j<N; j++) {
		i0 = ind[j-1];
		i1 = ind[j];

		yp = (red[j] - red[j-1])/(float)(i1-i0);
		y0 =  red[j-1];
		for(i=i0; i<i1; i++) { frgba[i][0] = y0 + yp * (float)(i-i0); }

		yp = (grn[j] - grn[j-1])/(float)(i1-i0);
		y0 =  grn[j-1];
		for(i=i0; i<i1; i++) { frgba[i][1] = y0 + yp * (float)(i-i0); }

		yp = (blu[j] - blu[j-1])/(float)(i1-i0);
		y0 =  blu[j-1];
		for(i=i0; i<i1; i++) { frgba[i][2] = y0 + yp * (float)(i-i0); }
	}
	      i0     = ind[N-1];
	frgba[i0][0] = red[N-1];
	frgba[i0][1] = grn[N-1];
	frgba[i0][2] = blu[N-1];

	if(m_iAlphaIsSet == 0) {
		for(i=0; i <= i0; i++) {
			frgba[i][3] = (float)0.1 * frgba[i][0] + (float)0.6 * frgba[i][1] + (float)0.3 * frgba[i][2];
		}

		m_iNalpNots = m_iNrgbNots;
		for(i=0; i < m_iNrgbNots; i++) {
			m_iALPind[j] = m_iRGBind[j];
			m_fALP[j]    = (float)0.1 * m_fRED[j] +(float)0.6 * m_fGRN[j] +(float)0.3 * m_fBLU[j];
		}
	}

//	FILE *fp;
//	fp = fopen("debug_SetLUTia.txt", "w");
//	for(i=0; i<256; i++) {
//		fprintf(fp, "%3d %f %f %f %f\n", i, frgba[i][0], frgba[i][1], frgba[i][2], frgba[i][3]);
//	}
//	fclose(fp);

/* This has been moved into drawscene
	if(m_bColorTableIsAvailable) { glColorTableEXT(DHP_TEX, GL_RGBA,256,GL_RGBA, GL_FLOAT,frgba); }
	glFlush();
	glFinish();
*/

	  for(i=0; i<256; i++) {
		  m_lrgba[i] =						  (unsigned long)(255.99*frgba[i][0]) +
						(unsigned long)256 * ((unsigned long)(255.99*frgba[i][1]) +
						(unsigned long)256 * ((unsigned long)(255.99*frgba[i][2]) +
						(unsigned long)256 *  (unsigned long)(255.99*frgba[i][3])		));
	  }


	m_iAlphaIsSet = 0;

	return S_OK;
}

STDMETHODIMP Chvr_work::SetLUTia(int N, int *ind, float *alpha)
{
	int i0, i1, i, j;
	float y0, yp;

	m_iNalpNots = N;
	for(j=0; j<N; j++) {
		m_iALPind[j] = ind[j];
		m_fALP[j]    = alpha[j];
	}

	for(j=1; j<N; j++) {
		i0 = ind[j-1];
		i1 = ind[j];

		yp = (alpha[j] - alpha[j-1])/(float)(i1-i0);
		y0 =  alpha[j-1];
		for(i=i0; i<i1; i++) { frgba[i][3] = y0 + yp * (float)(i-i0); }
	}
	      i0     =   ind[N-1];
	frgba[i0][3] = alpha[N-1];

	m_iAlphaIsSet = 1;

	return S_OK;
}

STDMETHODIMP Chvr_work::GetLUT32irgb(int *N, int *ind, float *red, float *grn, float *blu)
{
	int j;

	*N = m_iNrgbNots;
	for(j=0; j<m_iNrgbNots; j++) {
		ind[j] = m_iRGBind[j];
		red[j] = m_fRED[j];
		grn[j] = m_fGRN[j];
		blu[j] = m_fBLU[j];
	}


	return S_OK;
}

STDMETHODIMP Chvr_work::GetLUT32ia(int *N, int *ind, float *alp)
{
	int j;

	*N = m_iNalpNots;
	for(j=0; j<m_iNalpNots; j++) {
		ind[j] = m_iALPind[j];
		alp[j] = m_fALP[j];
	}

	return S_OK;
}


STDMETHODIMP Chvr_work::SetOption(wchar_t *wcInKey, wchar_t *wcInValue)
{
	// All of the gl stuff needs to be done by the rendering thread.
	// Hence: 1) copy input strings to static variables
	//        2) trigger the rendering thread to run DoSetOption
	//        3) wait for DoSetOption to finish
	wcscpy(wcKey, wcInKey);
	wcscpy(wcValue, wcInValue);
	iDoThis = 3;
	while(iDoThis == 3) { _sleep(10); }

	return S_OK;
}
void Chvr_work::DoSetOption()
{
	if(wcscmp(wcKey, L"UsePalet") == 0) {
		if(wcscmp(wcValue, L"TRUE") == 0) {
			m_bColorTableIsAvailable = TRUE;
			glColorTableEXT(DHP_TEX, GL_RGBA,256,GL_RGBA, GL_FLOAT,frgba);
		}
		else {
			m_bColorTableIsAvailable = FALSE;
		}
	}

	if(wcscmp(wcKey, L"TimersOn") == 0) { swscanf(wcValue, L"%d", &m_iTimersOn); }

	if(wcscmp(wcKey, L"TimerOutputFile") == 0) { swscanf(wcValue, L"%hs", m_sTimerFile); }

	if(wcscmp(wcKey, L"BufferedIO") == 0) { swscanf(wcValue, L"%d", &m_iBufferedIO); }

	if(wcscmp(wcKey, L"NRenderingBuffers") == 0) {
		swscanf(wcValue, L"%d", &m_iNRendBuffs);
		if(m_iNRendBuffs == 2) { glDrawBuffer(GL_BACK);  glReadBuffer(GL_BACK); }
		else                   { glDrawBuffer(GL_FRONT); glReadBuffer(GL_FRONT); }
	}
	if(wcscmp(wcKey, L"ShowColorBar") == 0) { swscanf(wcValue, L"%d", &m_iShowColorBar); }

	if(wcscmp(wcKey, L"ClearTicks" ) == 0) { m_iNTicks  = 0; }
	if(wcscmp(wcKey, L"ClearXTicks") == 0) { m_iXNTicks = 0; }
	if(wcscmp(wcKey, L"ClearYTicks") == 0) { m_iYNTicks = 0; }

	if(wcscmp(wcKey, L"AddTick") == 0) {
		swscanf(wcValue, L"%3d %hs", &m_iTicks[m_iNTicks], m_sTicks[m_iNTicks]);
		m_iNTicks++;
	}

	if(wcscmp(wcKey, L"AddXTick") == 0) {
		swscanf(wcValue, L"%3d %hs", &m_iXTicks[m_iXNTicks], m_sXTicks[m_iXNTicks]);
		m_iXNTicks++;
	}

	if(wcscmp(wcKey, L"AddYTick") == 0) {
		swscanf(wcValue, L"%3d %hs", &m_iYTicks[m_iYNTicks], m_sYTicks[m_iYNTicks]);
		m_iYNTicks++;
	}

	if(wcscmp(wcKey, L"BlockOnRender") == 0) { swscanf(wcValue, L"%d", &m_iBlockOnRender); }

	if(wcscmp(wcKey, L"Stats") == 0) { swscanf(wcValue, L"%hs", m_sStats); }

	if(wcscmp(wcKey, L"Yfile") == 0) { swscanf(wcValue, L"%hs", m_sTileO); }

	if(wcscmp(wcKey, L"NCacheMB") == 0) {
		int iMB_OLD = iNCacheMB;
		swscanf(wcValue, L"%hd", &iNCacheMB);
		if(iIO_Thread_Is_Up == 1) {
			SIZE_T  dwSizeOLD = (SIZE_T)iMB_OLD   * (SIZE_T)1024 * (SIZE_T)1024;
			SIZE_T  dwSizeNEW = (SIZE_T)iNCacheMB * (SIZE_T)1024 * (SIZE_T)1024;
			VirtualFree(BobCache, dwSizeOLD, MEM_DECOMMIT | MEM_RELEASE);
			BobCache = (unsigned char *)VirtualAlloc(NULL, dwSizeNEW,
													 MEM_COMMIT, PAGE_READWRITE);
			FlushCacheBuffers();
		}
	}
}

STDMETHODIMP Chvr_work::GetStatus(wchar_t **wcStatus)
{
	wchar_t wcStr[256];
	if(iDoThis == 0) { swprintf(wcStr, L"IDLE"); }
	else             { swprintf(wcStr, L"WORKING"); }

	long lAllocSize = (wcslen(wcStr)+1) * sizeof(wchar_t);
	*wcStatus = static_cast<wchar_t *>(CoTaskMemAlloc(lAllocSize));
	wcscpy(*wcStatus, wcStr);

	return S_OK;
}

STDMETHODIMP Chvr_work::GetPix(BinaryDataHVR *pData)
{
	int nbytes;

	// Set Rendering thread to run DoGetPix and wait for it to finish
	iDoThis = 5;
	while(iDoThis == 5) { _sleep(10); }

	nbytes = m_ShvrUse.iHeight * m_ShvrUse.iWidth * 3;
	pData->m_lSize = (long)nbytes;
	pData->m_pBuffer = static_cast<unsigned char *>(CoTaskMemAlloc(pData->m_lSize));
	memcpy(pData->m_pBuffer, ucGPix, sizeof(unsigned char) * nbytes);

	return S_OK;
}

STDMETHODIMP Chvr_work::PutPix(long lSize, unsigned char *cPix)
{
	memcpy(ucGPix, cPix, sizeof(unsigned char)*3*m_ShvrUse.iWidth*m_ShvrUse.iHeight);

	// Set Rendering thread to run DoGetPix and wait for it to finish
	iDoThis = 4;
	while(iDoThis == 4) { _sleep(10); }
	return S_OK;
}

void Chvr_work::DoGetPix()
{
//	if(m_iNRendBuffs == 2) { SwapBuffers(hDC); }
	glReadBuffer(GL_FRONT);
    glMatrixMode(GL_TEXTURE);
    glDisable(DHP_TEX);
    glDisable(GL_BLEND);
    glReadPixels( 0, 0, m_ShvrUse.iWidth, m_ShvrUse.iHeight, GL_RGB, GL_UNSIGNED_BYTE, ucGPix);
	glFlush();
	glFinish();
    glEnable(GL_BLEND);
    glEnable(DHP_TEX);
	if(m_iNRendBuffs == 2) { glReadBuffer(GL_BACK); }
//	if(m_iNRendBuffs == 2) { SwapBuffers(hDC); }
}

void Chvr_work::DoPutPix()
{
    glMatrixMode(GL_TEXTURE);
    glDisable(DHP_TEX);
    glDisable(GL_BLEND);
	glDrawPixels(m_ShvrUse.iWidth, m_ShvrUse.iHeight, GL_RGB,GL_UNSIGNED_BYTE, ucGPix);
	glFlush();
	glFinish();
    glEnable(GL_BLEND);
    glEnable(DHP_TEX);
	if(m_iNRendBuffs == 2) { SwapBuffers(hDC); }
}
